# Test that progress notification work
from copy import copy
from pathlib import Path
import time
import fasteners
from typing import List, Tuple, Union
from experimaestro import Task, Annotated, pathgenerator, progress, tqdm
from experimaestro.core.objects import logger
from experimaestro.notifications import LevelInformation
from experimaestro.scheduler import Job, Listener
from queue import Queue
from .utils import TemporaryExperiment


class ProgressingTask(Task):
    path: Annotated[Path, pathgenerator("progress.txt")]

    def execute(self):
        _progress = 0.0

        while True:
            time.sleep(1e-4)
            if self.path.is_file():
                with fasteners.InterProcessLock(self.path.with_suffix(".lock")):
                    _level, _progress, _desc = self.path.read_text().split(
                        " ", maxsplit=2
                    )
                    _progress = float(_progress)
                    _level = int(_level)
                    self.path.unlink()

                    if _progress > 0:
                        progress(_progress, level=_level, desc=_desc or None)
                    if _progress >= 0.99 and _level == 0:
                        break


def writeprogress(path: Path, progress, level=0, desc=None):
    """Write the progress to a file to see if this is correctly reported
    by the XPM server"""
    while True:
        time.sleep(5e-2)
        with fasteners.InterProcessLock(path.with_suffix(".lock")):
            if not path.is_file():
                path.write_text(f"{level} {progress:.3f} {desc if desc else ''}")
                break


class ProgressListener(Listener):
    def __init__(self):
        self.current = []
        self.progresses: Queue[List[LevelInformation]] = Queue()

    def job_state(self, job: Job):
        if (len(self.current) != len(job.progress)) or any(
            l1 != l2 for l1, l2 in zip(self.current, job.progress)
        ):
            logger.info("Got some progress: %s", job.progress)
            self.current = [copy(level) for level in job.progress]
            self.progresses.put(self.current)

        # if job.state.finished():
        #     self.progresses.put(-1 if job.state == JobState.DONE else -2)


def test_progress_basic():
    """Test that we get all the progress reports"""
    with TemporaryExperiment("progress-basic", maxwait=5, port=0) as xp:
        assert xp.server is not None
        assert xp.server.port > 0

        listener = ProgressListener()
        xp.scheduler.addlistener(listener)

        out = ProgressingTask().submit()
        path = out.path  # type: Path
        job = out.__xpm__.job

        logger.info("Waiting for job to start")
        while job.state.notstarted():
            time.sleep(1e-2)

        logger.info("Checking job progress")
        progresses = [i / 10.0 for i in range(11)]
        for v in progresses:
            writeprogress(path, v)
            if v < 1:
                info = listener.progresses.get()[0]
                logger.info("Got %s", info)
                assert info.progress == v


def test_progress_multiple():
    """Test that even with two schedulers, we get notified"""
    max_wait = 5

    with TemporaryExperiment(
        "progress-progress-multiple-1", maxwait=max_wait, port=0
    ) as xp1:
        assert xp1.server is not None
        assert xp1.server.port > 0

        listener1 = ProgressListener()
        xp1.scheduler.addlistener(listener1)

        out = ProgressingTask().submit()
        path = out.path  # type: Path
        job = out.__xpm__.job

        logger.info("Waiting for job to start (1)")
        while job.state.notstarted():
            time.sleep(1e-2)

        with TemporaryExperiment(
            "progress-progress-multiple-2",
            workdir=xp1.workdir,
            maxwait=max_wait,
            port=0,
        ) as xp2:
            assert xp2.server is not None
            assert xp2.server.port > 0
            listener2 = ProgressListener()
            xp2.scheduler.addlistener(listener2)

            out = ProgressingTask().submit()
            job = out.__xpm__.job  # type: CommandLineJob
            logger.info("Waiting for job to start (2)")
            while job.state.notstarted():
                time.sleep(1e-2)

            # Both schedulers should receive the job progress information
            logger.info("Checking job progress")
            progresses = [i / 10.0 for i in range(11)]
            for v in progresses:
                writeprogress(path, v)
                if v < 1:
                    assert listener1.progresses.get()[0].progress == v
                    assert listener2.progresses.get()[0].progress == v


NestedTasks = Tuple[str, Union[int, List["NestedTasks"]]]


class NestedProgressingTask(Task):
    PROGRESSES: NestedTasks = (
        "Task 1",
        [
            ("Task 1.1", 2),
            ("Task 1.2", [("Task 1.2.1", 3), ("Task 1.2.2", 4)]),
            ("Task 1.3", 1),
        ],
    )

    path: Annotated[Path, pathgenerator("progress.txt")]

    def execute(self):
        self._execute(NestedProgressingTask.PROGRESSES)

    def wait(self):
        while not self.path.exists():
            time.sleep(1e-4)
        self.path.unlink()

    def _execute(self, tasks: NestedTasks):
        self.wait()
        name, subtasks = tasks
        if isinstance(subtasks, list):
            for subtasks in tqdm(subtasks, desc=name, miniters=1, mininterval=0):
                self._execute(subtasks)
            self.wait()
        else:
            for _ in tqdm(range(subtasks), desc=name, miniters=1, mininterval=0):
                self.wait()
            self.wait()


def check_nested(
    path: Path, listener: ProgressListener, tasks: NestedTasks, level_ix=0
):
    name, subtasks = tasks

    def check_self(progress: float, nlevel: int = None):
        path.touch()
        levels = listener.progresses.get()
        logger.info(
            "Got %s (checking %d, %f, %d)",
            levels,
            level_ix,
            progress,
            nlevel or (level_ix + 1),
        )
        assert len(levels) == nlevel or (level_ix + 1)
        level = levels[level_ix]
        assert level.desc == name
        assert level.progress == progress

    if isinstance(subtasks, list):
        for ix, subsubtasks in enumerate(subtasks):
            logger.info(
                "Checking nested for %s (%d, %d/%d)", name, level_ix, ix, len(subtasks)
            )
            check_self(ix / len(subtasks), level_ix + 2)
            check_nested(path, listener, subsubtasks, level_ix + 1)
        check_self(1, level_ix + 2)
    else:
        for ix in range(subtasks):
            check_self(ix / float(subtasks))
        check_self(1)


def test_progress_nested():
    """Test that we get all the progress reports"""
    with TemporaryExperiment("progress-nested", maxwait=20, port=0) as xp:
        assert xp.server is not None
        assert xp.server.port > 0

        listener = ProgressListener()
        xp.scheduler.addlistener(listener)

        out = NestedProgressingTask().submit()
        job = out.__xpm__.job
        path = out.path  # type: Path

        logger.info("Waiting for job to start")
        while job.state.notstarted():
            time.sleep(1e-2)

        logger.info("Checking job progress")

        levels = listener.progresses.get()
        assert levels == [LevelInformation(0, None, 0.0)]
        check_nested(path, listener, NestedProgressingTask.PROGRESSES)
        logger.info("All good!")
        path.touch()



from typing import Callable, Dict, Optional
import uuid
import ipywidgets as widgets
from experimaestro import experiment
from IPython.display import display
from .jobs import jobmonitor  # noqa: F401


class serverwidget:
    TOKEN = uuid.uuid4().hex

    def __init__(
        self,
        name,
        *,
        port=12345,
        hook: Callable[["serverwidget"], None] = None,
        environment: Optional[Dict[str, str]] = None,
    ):
        self.name = name

        self.environment = environment or {}
        self.hook = hook
        self.port = port
        self.button = widgets.Button(description="Start the experimaestro server")
        self.output = widgets.Output()
        display(self.button, self.output)
        self.button.on_click(self.on_button_clicked)

        self.on_button_clicked(True)
        self.refresh()

    def refresh(self):
        self.output.clear_output()
        with self.output:
            if experiment.CURRENT:
                self.button.description = "Stop experimaestro server"
                print(  # noqa: T201
                    "Server started : "
                    f"http://localhost:{self.port}/auth?xpm-token={serverwidget.TOKEN}"
                )
            else:
                self.button.description = "Start experimaestro server"
                print("Server stopped")  # noqa: T201

    def on_button_clicked(self, b):
        with self.output:
            if experiment.CURRENT:
                try:
                    experiment.CURRENT.__exit__(None, None, None)
                except Exception:
                    print("Error while stopping experimaestro")  # noqa: T201
                self.current = experiment.CURRENT
            else:
                self.current = experiment(
                    self.name,
                    self.name,
                    host="localhost",
                    port=self.port,
                    token=serverwidget.TOKEN,
                ).__enter__()
                for key, value in self.environment.items():
                    self.current.setenv(key, value)
                if self.hook:
                    self.hook(self)

        self.refresh()



import threading


class OutputCaptureHandler:
    """Catpure the full output of a file stream"""

    def __init__(self, threaded=True):
        self._output = b""
        self.thread = None
        self.threaded = threaded

    def __call__(self, fp):
        if self.threaded:
            self.thread = threading.Thread(target=self.read, args=(fp,))
            self.thread.start()
        else:
            self.read(fp)

    @property
    def output(self):
        if self.thread:
            self.thread.join()
        return self._output

    def read(self, fp):
        while True:
            data = fp.read()
            if data == b"":
                break
            self._output += data



# flake8: noqa: F401

from .base import *
from .specs import (
    cpu,
    cuda_gpu,
    HostSpecification,
    CPUSpecification,
    CudaSpecification,
    HostRequirement,
    MatchRequirement,
)
from .registry import find_launcher, LauncherRegistry, YAMLDataClass
from .parser import parse



import importlib
import sys
from experimaestro import experiment

if __name__ == "__main__":
    wspath, module, functionname = sys.argv[1:]
    print("Importing", module)
    f = getattr(importlib.import_module(module), functionname)
    with experiment(wspath, "restart") as xp:
        f(xp)



from omegaconf import MISSING
from typing import Optional
import attr

try:
    from typing import dataclass_transform
except ImportError:
    from typing_extensions import dataclass_transform


@dataclass_transform(kw_only_default=True)
def configuration(*args, **kwargs):
    """Method to define keyword only dataclasses

    Configurations are keyword-only
    """

    return attr.define(*args, kw_only=True, slots=False, hash=True, eq=True, **kwargs)


@configuration()
class ConfigurationBase:
    """Base configuration for any experiment"""

    id: str = MISSING
    """ID of the experiment

    This ID is used by experimaestro when running as the experiment.
    """

    file: str = "experiment"
    """Relative path of the file containing a run function"""

    parent: Optional[str] = None
    """Relative path of a YAML file that should be merged"""

    title: str = ""
    """Short description of the experiment"""

    subtitle: str = ""
    """Allows to give some more details about the experiment"""

    paper: str = ""
    """Source paper for this experiment"""

    description: str = ""
    """Description of the experiment"""



from experimaestro import Config
from experimaestro.core.arguments import Annotated
from experimaestro.checkers import Choices
import pytest


def test_choices():
    """Test choices"""

    class TestChoices(Config):
        a: Annotated[str, Choices(["a", "b"])]

    TestChoices(a="a").__xpm__.validate()

    with pytest.raises((ValueError, KeyError)):
        TestChoices(a="c").__xpm__.validate()



import atexit
import shutil
import tempfile
from pathlib import Path
from subprocess import Popen, PIPE, run
import time
from rpyc.utils.server import OneShotServer
import rpyc
import threading
import logging

logger = logging.getLogger("rpyc")
logger.setLevel(logging.WARNING)


class client:
    def __init__(self, hostname: str, pythonpath: str, port: int = None):
        """[summary]

        Arguments:
            hostname -- The hostname to connect to
            pythonpath -- The remote python path
        """
        self.hostname = hostname
        self.pythonpath = pythonpath
        self.port = port

    # Constructs the SSH command line
    def ssh(self, *args):
        command = ["ssh"]
        command.extend(args)
        if self.port:
            command.extend(["-p", str(self.port)])
        command.append(self.hostname)
        return command

    def connect(self):
        # Get the local unix_path
        tmp = tempfile.TemporaryDirectory()
        tmpdirname = tmp.__enter__()
        local_unix_path = str(Path(tmpdirname) / "rpyc-client.sock")

        # Get the remote unix_path
        command = self.ssh()
        command.extend(["mktemp", "-d"])
        logger.debug("Runnning %s", command)
        p = run(command, capture_output=True)
        p.check_returncode()
        remote_unix_path = p.stdout.decode("utf-8").strip() + "/rpyc-server.sock"

        # Start server
        command = self.ssh(f"-L{local_unix_path}:{remote_unix_path}")
        command.extend(
            [
                self.pythonpath,
                "-m",
                "experimaestro",
                "rpyc-server",
                "--clean",
                remote_unix_path,
            ]
        )

        logger.debug("Runnning %s", command)
        process = Popen(command, stdout=PIPE)
        atexit.register(lambda process: process.kill(), process)

        # Wait for the server to be started
        process.stdout.readline()

        # Connect to server
        logger.info("Connecting to %s", local_unix_path)
        self.connection = rpyc.classic.unix_connect(local_unix_path)

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        logger.info("Closing connection")
        self.connection.close()


# --- Server part

server = None


class ClassicService(rpyc.core.service.ClassicService):
    """Full duplex master/slave service, i.e. both parties have full control
    over the other. Must be used by both parties."""

    __slots__ = "connected"

    def __init__(self):
        super().__init__()
        self.connected = False

    def on_connect(self, conn):
        self.connected = True
        super().on_connect(conn)

    def on_disconnect(self, conn):
        print("Disconnected")
        super().on_disconnect(conn)


def cleanup(path):
    logger.info("Cleaning up %s", path)
    path.unlink()
    path.parent.rmdir()


def start_server(unix_path, clean=None):
    service = ClassicService()
    server = OneShotServer(
        socket_path=str(unix_path), listener_timeout=1, service=service, logger=logger
    )

    def sayhello():
        while not server.active:
            time.sleep(0.01)
        print("HELLO", flush=True)
        logger.debug("Server started")

        time.sleep(5)
        if not service.connected:
            logger.warning("No inbound connection: stopping")
            server.close()

    threading.Thread(target=sayhello).start()
    if clean:
        atexit.register(cleanup, unix_path)
    server.start()


def stop_server():
    server.stop()



import logging
from typing import Any, ClassVar, Dict, List, Optional, Type, TypeVar
import typing
from yaml import Loader, Dumper, MappingNode

from experimaestro import typingutils

logger = logging.getLogger("xpm.yaml")


class YAMLException(Exception):
    def __init__(self, key: str, name: str, line: int, column: int, message: str):
        super().__init__(
            f"Exception while setting {key} in {name}:{line}:{column}: {message}"
        )


def check_type(object, typehint):
    typehint = typingutils.get_type(typehint)
    if list_type := typingutils.get_list(typehint):
        assert isinstance(object, list), f"{object} is not a list"
        for el in object:
            check_type(el, list_type)
    elif dict_type := typingutils.get_dict(typehint):
        assert isinstance(object, dict)
        for key, value in object.items():
            check_type(key, dict_type[0])
            check_type(value, dict_type[1])
    else:
        assert isinstance(
            object, typehint
        ), f"{type(object)} is not a subtype of {typehint}"


def register_yaml(
    cls, yaml_loader: Type[Loader], yaml_dumper: Optional[Type[Dumper]] = None
):
    yaml_tag = getattr(cls, "yaml_tag", None)
    if yaml_tag is None and getattr(cls, "from_yaml", None):
        cls.yaml_tag = f"!{cls.__module__}.{cls.__qualname__}"
        logger.debug("Registering constructor %s / cls %s", cls.yaml_tag, cls.from_yaml)
        yaml_loader.add_constructor(cls.yaml_tag, cls.from_yaml)
        if yaml_dumper and (to_yaml := getattr(cls, "to_yaml", None)):
            logger.debug("Registering YAML dumper for %s", cls)
            yaml_dumper.add_representer(cls, to_yaml)
        return True

    # None when no from_yaml
    return yaml_tag is not None


class YAMLDataClass:
    """Base class for dataclass driven YAML objects"""

    __dataclass_fields__: ClassVar[Dict[str, Any]]
    yaml_tag: ClassVar[str]

    @classmethod
    def to_yaml(cls, dumper: Dumper, data):
        return dumper.represent_dict(
            {key: getattr(data, key) for key, value in cls.__dataclass_fields__.items()}
        )

    @classmethod
    def from_yaml(cls, loader: Loader, node):
        kwargs = {}
        for key, value in node.value:
            try:
                assert isinstance(key.value, str)
                assert (
                    key.value in cls.__dataclass_fields__
                ), f"{key.value} is not a valid field for {cls}"

                v = loader.construct_object(value, deep=True)

                fieldtype = cls.__dataclass_fields__[key.value].type
                if typingutils.is_annotated(fieldtype):
                    initializers = [
                        x
                        for x in typingutils.get_args(fieldtype)
                        if isinstance(x, Initialize)
                    ]
                    assert (
                        len(initializers) <= 1
                    ), "Too many initializers for this field"
                    if initializers:
                        v = initializers[0].fn(v)

                # Cast if needed
                origin = typingutils.get_type(fieldtype)
                origin = typing.get_origin(origin) or origin
                if origin and not isinstance(v, origin):
                    v = origin(v)

                check_type(v, fieldtype)
                kwargs[key.value] = v
            except YAMLException:
                raise
            except Exception as e:
                raise YAMLException(
                    key.value,
                    node.start_mark.name,
                    node.start_mark.line,
                    node.start_mark.column,
                    str(e),
                )

        return cls(**kwargs)


T = TypeVar("T")


class YAMLList(List[T]):
    @classmethod
    def from_yaml(cls, loader, node):
        array = [loader.construct_object(el, deep=True) for el in node.value]
        ctype = cls.get_ctype()
        for el in array:
            assert isinstance(el, ctype)

        return cls(array)

    @classmethod
    def get_ctype(cls):
        (list_type,) = [
            p for p in cls.__orig_bases__ if typing.get_origin(p) is YAMLList
        ]
        (list_element_type,) = typingutils.get_args(list_type)
        return list_element_type


class YAMLDict(Dict[str, T]):
    @classmethod
    def from_yaml(cls, loader, node):
        map = {}
        ctype = cls.get_ctype()

        for node_key, node_value in node.value:
            key = loader.construct_scalar(node_key)
            value = loader.construct_object(node_value, deep=True)
            assert isinstance(value, ctype), f"{value} is not of type {ctype}"
            map[key] = value

        print(map)
        return cls(**map)

    @classmethod
    def get_ctype(cls):
        (list_type,) = [
            p for p in cls.__orig_bases__ if typing.get_origin(p) is YAMLDict
        ]
        (value_type,) = typingutils.get_args(list_type)
        return value_type


class Initialize:
    def __init__(self, fn):
        self.fn = fn


def add_path_resolvers(
    loader: Type[Loader],
    path: List[Optional[str]],
    cls: typing.Type,
    dumper: Optional[Type[Dumper]] = None,
):
    cls = typingutils.get_type(cls) or cls

    if list_type := typingutils.get_list(cls):
        add_path_resolvers(loader, path + [None], list_type, dumper=dumper)

    elif dict_type := typingutils.get_dict(cls):
        key_type, value_type = dict_type
        assert issubclass(key_type, str), f"Key of dict must be a string ({key_type})"
        add_path_resolvers(loader, path + [None], value_type, dumper=dumper)

    else:
        if register_yaml(cls, loader, yaml_dumper=dumper):
            logger.debug(f"Adding {cls.yaml_tag} -> {path}")

            if issubclass(cls, YAMLDataClass):
                # Add others
                loader.add_path_resolver(cls.yaml_tag, path, dict)
                for field in cls.__dataclass_fields__.values():
                    # logger.debug("Processing %s", field)
                    field_type = typingutils.get_type(field.type) or field.type
                    add_path_resolvers(
                        loader, path + [field.name], field_type, dumper=dumper
                    )

            elif issubclass(cls, YAMLDict):
                loader.add_path_resolver(cls.yaml_tag, path + [None], dict)
                add_path_resolvers(
                    loader, path + [None], cls.get_ctype(), dumper=dumper
                )

            elif issubclass(cls, YAMLList):
                loader.add_path_resolver(cls.yaml_tag, path, list)
                add_path_resolvers(
                    loader, path + [None], cls.get_ctype(), dumper=dumper
                )



from pathlib import Path
from typing import Optional, Union
from experimaestro import Config
from experimaestro.core.context import SerializedPath
from experimaestro.core.objects import ConfigInformation
from huggingface_hub import ModelHubMixin, hf_hub_download, snapshot_download
import os


class ExperimaestroHFHub(ModelHubMixin):
    """Defines models that can be uploaded/downloaded from the Hub"""

    def __init__(self, config: Config, variant: Optional[str] = None):
        self.config = config
        self.variant = variant

    def _save_pretrained(self, save_directory: Union[str, Path]):
        save_directory = Path(save_directory)
        if self.variant:
            save_directory = save_directory / self.variant
            save_directory.mkdir()
        assert self.config is not None
        self.config.__xpm__.serialize(save_directory)

    @classmethod
    def _from_pretrained(
        cls,
        model_id,
        revision,
        cache_dir,
        force_download,
        proxies,
        resume_download,
        local_files_only,
        token,
        *,
        variant: Optional[str] = None,
        as_instance: bool = False,
        **model_kwargs,
    ):
        if os.path.isdir(model_id):
            save_directory = Path(model_id)

            def data_loader(path: Path):
                if variant:
                    return save_directory / path / variant
                return save_directory / path

        else:

            def data_loader(s_path: Union[Path, str, SerializedPath]):
                if not isinstance(s_path, SerializedPath):
                    s_path = SerializedPath(Path(s_path), False)
                path = s_path.path

                # Folder
                if s_path.is_folder:
                    hf_path = snapshot_download(
                        repo_id=model_id,
                        allow_patterns=f"{s_path.path}/**",
                        revision=revision,
                        cache_dir=cache_dir,
                        proxies=proxies,
                        resume_download=resume_download,
                        token=token,
                        local_files_only=local_files_only,
                    )
                    return Path(hf_path) / path

                hf_path = Path(
                    hf_hub_download(
                        repo_id=model_id,
                        filename=str(path if variant is None else Path(variant) / path),
                        revision=revision,
                        cache_dir=cache_dir,
                        force_download=force_download,
                        proxies=proxies,
                        resume_download=resume_download,
                        token=token,
                        local_files_only=local_files_only,
                    )
                )
                return hf_path

        return ConfigInformation.deserialize(data_loader, as_instance=as_instance)



import logging
import pytest
import os
import shutil


@pytest.fixture(scope="session")
def xpmdirectory(tmp_path_factory):
    """Sets a temporary main directory"""
    workdir = tmp_path_factory.mktemp("workdir")
    logging.info("Main XPM directory is %s", workdir)
    os.environ["XPM_WORKDIR"] = str(workdir)

    yield workdir

    if os.environ.get("XPM_KEEPWORKDIR", False) == "1":
        logging.warning("NOT Removing %s" % workdir)
    else:
        shutil.rmtree(workdir)


# Sets a flag
def pytest_configure(config):
    import sys

    sys._called_from_test = True


def pytest_unconfigure(config):
    import sys  # This was missing from the manual

    del sys._called_from_test



# Generate a nice documentation for experimaestro Configuration
# objects

from inspect import Parameter, Signature
from pathlib import Path
from sphinx.util.fileutil import copy_asset
import experimaestro
from typing import Any, Dict, List, Optional, Tuple
from docutils import nodes
from sphinx.application import Sphinx
from sphinx import addnodes
from sphinx.ext.autodoc import ClassDocumenter, Documenter, restify
from sphinx.locale import _, __
from sphinx.util import inspect, logging
from sphinx.domains.python import (
    PyClasslike,
    PyAttribute,
    PyObject,
    directives,
    desc_signature,
    _parse_annotation,
)
from sphinx.util.typing import OptionSpec
from docutils.statemachine import StringList
import logging
import re

from experimaestro import Config, Task

# See Sphinx documentation on developping extensions
# https://www.sphinx-doc.org/en/master/extdev/index.html#dev-extensions

logger = logging.getLogger(__name__)


class XPMConfigDirective(PyClasslike):
    """Directive for XPM configurations/tasks"""

    option_spec: OptionSpec = PyClasslike.option_spec.copy()

    option_spec.update(
        {
            "task": directives.unchanged,
        }
    )

    def get_signature_prefix(self, sig: str) -> List[nodes.Node]:
        """May return a prefix to put before the object name in the
        signature.
        """
        if self.options.get("task") is not None:
            return [nodes.Text("XPM Task")]
        return [nodes.Text("XPM Config")]


class XPMParamDirective(PyAttribute):
    """Description of an attribute."""

    option_spec: OptionSpec = PyAttribute.option_spec.copy()
    option_spec.update(
        {
            "constant": directives.unchanged,
            "generated": directives.unchanged,
        }
    )

    def handle_signature(self, sig: str, signode: desc_signature) -> Tuple[str, str]:
        fullname, prefix = super().handle_signature(sig, signode)

        if self.options.get("constant") is not None:
            signode += addnodes.desc_annotation("constant", nodes.Text("constant"))
        if self.options.get("generated") is not None:
            signode += addnodes.desc_annotation("generated", nodes.Text("generated"))

        return fullname, prefix


def getxpminfo(obj):
    """Returns the XPM type information for a configuration object"""
    config: Config = obj
    return config.__getxpmtype__()


class ConfigDocumenter(ClassDocumenter):
    """Documenter for experimaestro configurations and tasks"""

    objtype = "xpmconfig"
    directivetype = "xpmconfig"

    # We need to have a higher priority than class documenter to get called
    priority = 10 + ClassDocumenter.priority

    option_spec = dict(ClassDocumenter.option_spec)

    @classmethod
    def can_document_member(cls, member: Any, *args) -> bool:
        can_document = inspect.isclass(member) and issubclass(member, Config)
        return can_document

    def add_directive_header(self, sig: str) -> None:
        super().add_directive_header(sig)

    def get_object_members(self, want_all: bool):  # -> Tuple[bool, ObjectMembers]:
        r = super().get_object_members(want_all)
        return r

    def _get_signature(
        self,
    ) -> Tuple[Optional[Any], Optional[str], Optional[Signature]]:
        r = super()._get_signature()

        xpminfo = getxpminfo(self.object)

        p = []
        for name, argument in xpminfo.arguments.items():
            if not (argument.constant or argument.generator is not None):
                p.append(Parameter(name, Parameter.KEYWORD_ONLY))
        s = Signature(p)

        return r[0], r[1], s

    @staticmethod
    def formatDefault(value) -> str:
        if isinstance(value, Config):
            objecttype = value.__xpmtype__.objecttype
            params = ", ".join(
                [f"{key}={value}" for key, value in value.__xpm__.values.items()]
            )
            # It would be possible to do better... if not
            return f"{objecttype.__module__}.{objecttype.__qualname__}({params})"

        return str(value)

    def add_directive_header(self, sig: str) -> None:
        sourcename = self.get_sourcename()
        xpminfo = getxpminfo(self.object)

        # Mostly copied from ClassDocumenter but adapted (no other choice)
        # FIXME: would be better to be a plugin

        if self.doc_as_attr:
            self.directivetype = "attribute"

        # Skip ClassDocumenter
        Documenter.add_directive_header(self, sig)

        if self.analyzer and ".".join(self.objpath) in self.analyzer.finals:
            self.add_line("   :final:", sourcename)

        canonical_fullname = self.get_canonical_fullname()
        if (
            not self.doc_as_attr
            and canonical_fullname
            and self.fullname != canonical_fullname
        ):
            self.add_line("   :canonical: %s" % canonical_fullname, sourcename)

        # Our specific code
        if issubclass(self.object, Task):
            self.add_line(f"   :task:", sourcename)

        # add inheritance info, if wanted
        if not self.doc_as_attr and self.options.show_inheritance:
            if inspect.getorigbases(self.object):
                # A subclass of generic types
                # refs: PEP-560 <https://www.python.org/dev/peps/pep-0560/>
                bases = list(self.object.__orig_bases__)
            elif hasattr(self.object, "__bases__") and len(self.object.__bases__):
                # A normal class
                bases = list(self.object.__bases__)
            else:
                bases = []

            self.env.events.emit(
                "autodoc-process-bases", self.fullname, self.object, self.options, bases
            )

            if self.config.autodoc_typehints_format == "short":
                base_classes = [restify(cls, "smart") for cls in bases]
            else:
                base_classes = [restify(cls) for cls in bases]

            sourcename = self.get_sourcename()
            self.add_line("", sourcename)
            self.add_line("   " + _("Bases: %s") % ", ".join(base_classes), sourcename)

        # Adds return type if different
        if xpminfo.returntype != xpminfo.objecttype:
            self.add_line("", sourcename)
            self.add_line(
                "   " + _("Submit type: %s") % restify(xpminfo.returntype), sourcename
            )

            # annotations = _parse_annotation(str(xpminfo.returntype), self.env)

    def add_content(
        self, more_content: Optional[StringList], no_docstring: bool = False
    ) -> None:

        xpminfo = getxpminfo(self.object)
        source_name = self.get_sourcename()

        super().add_content(more_content)

        for name, argument in xpminfo.arguments.items():
            typestr = argument.type.name()

            self.add_line(f".. py:xpmparam:: {name}", source_name)
            self.add_line(f"   :type: {typestr}", source_name)

            if argument.default is not None:
                self.add_line(
                    f"   :value: {ConfigDocumenter.formatDefault(argument.default)}",
                    source_name,
                )
            if argument.generator:
                self.add_line(f"   :generated:", source_name)
            elif argument.constant:
                self.add_line(f"   :constant:", source_name)

            # self.add_line("", source_name)
            if argument.help:
                self.add_line("", source_name)
                self.add_line("    " + argument.help, source_name)
            self.add_line("", source_name)


RE_ATTRIBUTE = re.compile(r"""^.. attribute:: (.*)""")


def _process_docstring(
    app: Sphinx, what: str, name: str, obj: Any, options: Any, lines: List[str]
) -> None:
    """Removes already documented (by XPM) attributes from docstring

    Comes after napoleon (lesser priority) so benefit from a standardized
    output
    """
    if obj is not None and ConfigDocumenter.can_document_member(obj):
        xpminfo = getxpminfo(obj)
        names = set(xpminfo.arguments.keys())

        newlines = []
        skip = False
        for line in lines:
            if m := RE_ATTRIBUTE.match(line):
                if m.group(1) in names:
                    skip = True
            elif skip:
                if line.startswith(".."):
                    skip = False

            if not skip:
                newlines.append(line)

        lines[:] = newlines[:]


def copy_asset_files(app, exc):
    import os

    assetspath = Path(__file__).parent / "static"
    if exc is None:  # build succeeded
        for path in assetspath.glob("*.css"):
            copy_asset(str(path.absolute()), os.path.join(app.outdir, "_static"))


def setup(app: Sphinx) -> Dict[str, Any]:
    """Setup experimaestro for Sphinx documentation"""

    # Adds our stylesheet
    app.connect("build-finished", copy_asset_files)
    app.add_css_file("experimaestro.css")

    # We need autodoc
    app.setup_extension("sphinx.ext.autodoc")

    # Defines our extensions (documenter, domains)
    app.add_autodocumenter(ConfigDocumenter)
    app.add_directive("py:xpmconfig", XPMConfigDirective)

    pydomain = app.registry.domains["py"]
    pydomain.object_types["xpmconfig"] = pydomain.object_types["class"]
    # pydomain.add_object_type('py:xpmconfig', pydomain.object_types["class"])

    app.add_directive("py:xpmparam", XPMParamDirective)
    app.connect("autodoc-process-docstring", _process_docstring, 600)

    return {"version": experimaestro.__version__, "parallel_read_safe": True}



from dataclasses import dataclass, field
from functools import cached_property
from typing import Dict, List, Optional
from experimaestro.launcherfinder import (
    LauncherConfiguration,
    LauncherRegistry,
    HostRequirement,
)
from experimaestro.launcherfinder.registry import CPU, GPUList, YAMLDataClass
from experimaestro.launcherfinder.specs import (
    HostSpecification,
)
from experimaestro.scriptbuilder import PythonScriptBuilder
from . import Launcher


class DirectLauncher(Launcher):
    def scriptbuilder(self):
        return PythonScriptBuilder()

    @staticmethod
    def init_registry(registry: LauncherRegistry):
        registry.register_launcher("local", DirectLauncherConfiguration)

    def __str__(self):
        return f"DirectLauncher({self.connector})"


@dataclass
class DirectLauncherConfiguration(YAMLDataClass, LauncherConfiguration):
    connector: str = "connector"
    cpu: CPU = field(default_factory=CPU)
    gpus: GPUList = field(default_factory=GPUList)
    tokens: Optional[Dict[str, int]] = None
    tags: List[str] = field(default_factory=lambda: [])
    weight: int = 0
    disable: bool = False

    @cached_property
    def spec(self) -> HostSpecification:
        return HostSpecification(cpu=self.cpu.to_spec(), cuda=self.gpus.to_spec())

    def get(
        self, registry: LauncherRegistry, requirement: "HostRequirement"
    ) -> Optional[Launcher]:
        if requirement.match(self.spec):
            launcher = DirectLauncher(connector=registry.getConnector(self.connector))
            if self.tokens:
                for token_identifier, count in self.tokens.items():
                    token = registry.getToken(token_identifier)
                    # TODO: handle the case where this is not a CounterToken
                    launcher.addListener(
                        lambda job: job.dependencies.add(token.dependency(count))
                    )
            return launcher

        return None



# --- Parse specifications from strings

from functools import reduce
from arpeggio import (
    ZeroOrMore,
    OneOrMore,
    OrderedChoice,
    RegExMatch,
    ParserPython,
    StrMatch,
    Optional,
    PTNodeVisitor,
    visit_parse_tree,
    EndOfFile,
)
from . import specs

# --- Grammar


class SuppressStrMatch(StrMatch):
    suppress = True


def mem_spec():
    return "mem", "=", RegExMatch(r"\d+(G|M)?")


def cores_spec():
    return "cores", "=", RegExMatch(r"\d+")


def multiplier():
    return "*", RegExMatch(r"\d+")


def cuda_specs():
    return ZeroOrMore(OrderedChoice([mem_spec]), sep=",")


def cuda():
    return "cuda", "(", cuda_specs, ")", Optional(multiplier)


def cpu_specs():
    return ZeroOrMore(OrderedChoice([mem_spec, cores_spec]), sep=",")


def cpu():
    return "cpu", "(", cpu_specs, ")"


def duration():
    return "duration", "=", RegExMatch(r"\d+"), RegExMatch(r"h(ours)?|d(ays)?")


def one_spec():
    return OneOrMore(OrderedChoice([duration, cuda, cpu]), sep="&")


def grammar():
    return OneOrMore(one_spec, sep="|"), EndOfFile()


# ---- Visitor


class Visitor(PTNodeVisitor):
    def visit_grammar(self, node, children):
        return [child for child in children]

    def visit_one_spec(self, node, children):
        return reduce(lambda x, el: x & el, children)

    def visit_duration(self, node, children):
        return specs.duration(" ".join(children))

    def visit_cuda(self, node, children):
        if len(children) > 1:
            return specs.cuda_gpu(**children[0]) * int(children[1])
        return specs.cuda_gpu(**children[0])

    def visit_cpu(self, node, children):
        return specs.cpu(**children[0])

    def visit_specs(self, node, children):
        a = children[0]
        for c in children[1:]:
            a.update(c)
        return a

    visit_cuda_specs = visit_specs
    visit_cpu_specs = visit_specs

    def visit_mem_spec(self, node, children):
        return {"mem": node.value}

    def visit_cores_spec(self, node, children):
        return {"cores": int(node.value)}


def parse(expr: str):
    parser = ParserPython(grammar, syntax_classes={"StrMatch": SuppressStrMatch})
    parse_tree = parser.parse(expr)
    return visit_parse_tree(parse_tree, Visitor(debug=False))



from datetime import datetime
import logging
import asyncio
import platform
import socket
import uuid
from experimaestro.scheduler.base import Job
import pkg_resources
import http
import threading
from typing import Optional, Tuple
from experimaestro.scheduler import Scheduler, Listener as BaseListener
from experimaestro.scheduler.services import Service, ServiceListener
from experimaestro.settings import ServerSettings
from flask import Flask, Request, Response
from flask import request, redirect
from flask_socketio import SocketIO, emit, ConnectionRefusedError
import requests


def formattime(v: Optional[float]):
    if not v:
        return ""

    return datetime.fromtimestamp(v).isoformat()


def progress_state(job: Job):
    return [
        {"level": o.level, "progress": o.progress, "desc": o.desc} for o in job.progress
    ]


def job_details(job):
    return {
        "jobId": job.identifier,
        "taskId": job.name,
        "locator": str(job.jobpath),
        "status": job.state.name.lower(),
        "start": formattime(job.starttime),
        "end": formattime(job.endtime),
        "submitted": formattime(job.submittime),
        "tags": list(job.tags.items()),
        "progress": progress_state(job),
    }


def job_create(job: Job):
    return {
        "jobId": job.identifier,
        "taskId": job.name,
        "locator": str(job.jobpath),
        "status": job.state.name.lower(),
        "tags": list(job.tags.items()),
        "progress": progress_state(job),
    }


class Listener(BaseListener, ServiceListener):
    def __init__(self, scheduler: Scheduler, socketio):
        self.scheduler = scheduler
        self.socketio = socketio
        self.scheduler.addlistener(self)
        self.services = {}
        for service in self.scheduler.xp.services.values():
            self.service_add(service)

    def job_submitted(self, job):
        self.socketio.emit("job.add", job_create(job))

    def job_state(self, job):
        self.socketio.emit(
            "job.update",
            {
                "jobId": job.identifier,
                "status": job.state.name.lower(),
                "progress": progress_state(job),
            },
        )

    def service_add(self, service: Service):
        service.add_listener(self)
        self.services[service.id] = service
        self.socketio.emit(
            "service.add",
            {
                "id": service.id,
                "description": service.description(),
                "state": service.state.name,
            },
        )

    def service_state_changed(self, service: Service):
        self.socketio.emit("service.update", {"state": service.state.name})


MIMETYPES = {
    "html": "text/html",
    "map": "text/plain",
    "txt": "text/plain",
    "ico": "image/x-icon",
    "png": "image/png",
    "css": "text/css",
    "js": "application/javascript",
    "json": "application/json",
    "eot": "font/vnd.ms-fontobject",
    "woff": "font/woff",
    "woff2": "font/woff2",
    "ttf": "font/ttf",
}


def proxy_response(base_url: str, request: Request, path: str):
    # Whitelist a few headers to pass on
    request_headers = {}
    for key, value in request.headers.items():
        request_headers[key] = value

    if request.query_string:
        path = f"""{path}?{request.query_string.decode("utf-8")}"""

    data = None
    if request.method == "POST":
        data = request.get_data()

    response = requests.request(
        request.method,
        f"{base_url}{path}",
        data=data,
        stream=True,
        headers=request_headers,
    )
    headers = {}
    for key, value in response.headers.items():
        headers[key] = value

    flask_response = Response(
        response=response.raw.read(),
        status=response.status_code,
        headers=headers,
        content_type=response.headers["content-type"],
    )
    return flask_response


def start_app(server: "Server"):
    logging.debug("Starting Flask server...")
    app = Flask("experimaestro")

    logging.debug("Starting Flask server (SocketIO)...")
    socketio = SocketIO(app, path="/api", async_mode="gevent")
    listener = Listener(server.scheduler, socketio)

    logging.debug("Starting Flask server (setting up socketio)...")

    @socketio.on("connect")
    def handle_connect():
        if server.token != request.cookies.get("experimaestro_token", None):
            raise ConnectionRefusedError("invalid token")

    @socketio.on("refresh")
    def handle_refresh():
        for job in listener.scheduler.jobs.values():
            emit("job.add", job_create(job))

    @socketio.on("job.details")
    def handle_details(jobid):
        emit("job.update", job_details(listener.scheduler.jobs[jobid]))

    @socketio.on("services")
    def handle_services_list():
        for service in listener.services.values():
            emit(
                "service.add",
                {
                    "id": service.id,
                    "description": service.description(),
                    "state": service.state.name,
                },
            )

    @socketio.on("job.kill")
    def handle_job_kill(jobid: str):
        job = server.scheduler.jobs[jobid]
        future = asyncio.run_coroutine_threadsafe(
            job.aio_process(), server.scheduler.loop
        )
        process = future.result()
        if process is not None:
            process.kill()

    logging.debug("Starting Flask server (setting up routes)...")

    @app.route("/services/<path:path>", methods=["GET", "POST"])
    def route_service(path):
        service, *path = path.split("/", 1)
        if not path:
            return redirect(f"/services/{service}/", http.HTTPStatus.PERMANENT_REDIRECT)

        service = server.scheduler.xp.services.get(service, None)
        if service is None:
            return Response(f"Service {service} not found", http.HTTPStatus.NOT_FOUND)

        base_url = service.get_url()
        return proxy_response(base_url, request, path[0] if path else "/")

    @app.route("/notifications/<jobid>/progress")
    def notifications_progress(jobid):
        level = int(request.args.get("level", 0))
        progress = float(request.args.get("progress", 0.0))

        try:
            server.scheduler.jobs[jobid].set_progress(
                level,
                progress,
                request.args.get("desc", None),
            )
        except KeyError:
            # Just ignore
            pass
        return Response("", http.HTTPStatus.OK)

    @app.route("/")
    def route_root():
        if server.token == request.cookies.get("experimaestro_token", None):
            return redirect("/index.html", 302)
        return redirect("/login.html", 302)

    @app.route("/auth")
    def route_auth():
        if token := request.args.get("xpm-token", None):
            if server.token == token:
                resp = redirect("/index.html", 302)
                resp.set_cookie("experimaestro_token", token)
                return resp
        return redirect("/login.html", 302)

    @app.route("/stop")
    def route_stop():
        if (server.token == request.args.get("xpm-token", None)) or (
            server.token == request.cookies.get("experimaestro_token", None)
        ):
            socketio.stop()
            return Response(status=http.HTTPStatus.ACCEPTED)
        return Response(status=http.HTTPStatus.UNAUTHORIZED)

    @app.route("/<path:path>")
    def static_route(path):
        if token := request.form.get("experimaestro_token", None):
            if server.token == token:
                request.cookies["experimaestro_token"] = token

        if path == "index.html":
            if server.token != request.cookies.get("experimaestro_token", None):
                return redirect("/login.html", code=302)

        datapath = "data/%s" % path
        logging.debug("Looking for %s", datapath)
        if pkg_resources.resource_exists("experimaestro.server", datapath):
            mimetype = MIMETYPES[datapath.rsplit(".", 1)[1]]
            content = pkg_resources.resource_string("experimaestro.server", datapath)
            return Response(content, mimetype=mimetype)
        return Response("Page not found", status=404)

    # Start the app
    if server.port is None or server.port == 0:
        logging.info("Searching for an available port")
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind(("", 0))
        server.port = sock.getsockname()[1]
        sock.close()

    logging.info(
        "Web server started on http://%s:%d/auth?xpm-token=%s",
        server.host,
        server.port,
        server.token,
    )

    server.instance = socketio
    with server.cv_running:
        server.running = True
        server.cv_running.notify()
    socketio.run(
        app,
        host=server.host,
        port=server.port,
        debug=False,
        use_reloader=False,
    )
    logging.info("Web server stopped")


class Server:
    def __init__(self, scheduler: Scheduler, settings: ServerSettings):
        if settings.autohost == "fqdn":
            settings.host = socket.getfqdn()
            logging.info("Auto host name (fqdn): %s", settings.host)
        elif settings.autohost == "name":
            settings.host = platform.node()
            logging.info("Auto host name (name): %s", settings.host)

        if settings.host is None or settings.host == "127.0.0.1":
            self.bindinghost = "127.0.0.1"
        else:
            self.bindinghost = "0.0.0.0"

        self.host = settings.host or "127.0.0.1"
        self.port = settings.port
        self.scheduler = scheduler
        self.token = settings.token or uuid.uuid4().hex
        self.instance = None
        self.running = False
        self.cv_running = threading.Condition()

    def getNotificationSpec(self) -> Tuple[str, str]:
        """Returns a tuple (server ID, server URL)"""
        return (
            f"""{self.host}_{self.port}.url""",
            f"""http://{self.host}:{self.port}/notifications""",
        )

    def stop(self):
        if self.instance:
            try:
                requests.get(
                    f"http://{self.host}:{self.port}/stop?xpm-token={self.token}"
                )
            except requests.exceptions.ConnectionError:
                # This is expected
                pass

    def start(self):
        """Start the websocket server in a new process process"""
        logging.info("Starting the web server")

        # Avoids clutering
        logging.getLogger("geventwebsocket.handler").setLevel(logging.WARNING)

        self.thread = threading.Thread(target=start_app, args=(self,)).start()

        # Wait until we really started
        while True:
            with self.cv_running:
                if self.running:
                    break



# --- Task and types definitions

from pathlib import Path
import pytest
import logging
from experimaestro import Config, deprecate, Task, Param
from experimaestro.scheduler.workspace import RunMode
from experimaestro.tools.jobs import fix_deprecated
from experimaestro.scheduler import FailedExperiment, JobState
from experimaestro import SubmitHook, Job, Launcher, LightweightTask

from .utils import TemporaryDirectory, TemporaryExperiment

from .tasks.all import (
    Concat,
    ForeignTaskA,
    Fail,
    FailConsumer,
    SetUnknown,
    Method,
    Say,
    SimpleTask,
    CacheConfig,
    CacheConfigTask,
)
from . import restart
from .definitions_types import IntegerTask, FloatTask


def test_task_types():
    with TemporaryExperiment("simple"):
        assert IntegerTask(value=5).submit().__xpm__.job.wait() == JobState.DONE
        assert FloatTask(value=5.1).submit().__xpm__.job.wait() == JobState.DONE


def test_simple_task():
    with TemporaryDirectory(prefix="xpm", suffix="helloworld") as workdir:
        assert isinstance(workdir, Path)
        with TemporaryExperiment("helloworld", workdir=workdir, maxwait=20):
            # Submit the tasks
            hello = Say(word="hello").submit()
            world = Say(word="world").submit()

            # Concat will depend on the two first tasks
            concat = Concat(strings=[hello, world]).submit()

        assert concat.__xpm__.job.state == JobState.DONE
        assert Path(concat.stdout()).read_text() == "HELLO WORLD\n"


def test_not_submitted():
    """A not submitted task should not be accepted as an argument"""
    with TemporaryExperiment("helloworld", maxwait=2):
        hello = Say(word="hello")
        with pytest.raises(ValueError):
            Concat(strings=[hello])


def test_fail_simple():
    """Failing task... should fail"""
    with pytest.raises(FailedExperiment):
        with TemporaryExperiment("failing", maxwait=20):
            fail = Fail().submit()
            fail.touch()


def test_foreign_type():
    """When the argument real type is in an non imported module"""
    with TemporaryExperiment("foreign_type", maxwait=2):
        # Submit the tasks
        from .tasks.foreign import ForeignClassB2

        b = ForeignClassB2(x=1, y=2)
        a = ForeignTaskA(b=b).submit()

        assert a.__xpm__.job.wait() == JobState.DONE
        assert a.stdout().read_text().strip() == "1"


def test_fail_dep():
    """Failing task... should cancel dependent"""
    with pytest.raises(FailedExperiment):
        with TemporaryExperiment("failingdep"):
            fail = Fail().submit()
            dep = FailConsumer(fail=fail).submit()
            fail.touch()

    assert fail.__xpm__.job.state == JobState.ERROR
    assert dep.__xpm__.job.state == JobState.ERROR


def test_unknown_attribute():
    """No check when setting attributes while executing"""
    with TemporaryExperiment("unknown"):
        method = SetUnknown().submit()

    assert method.__xpm__.job.wait() == JobState.DONE


def test_function():
    with TemporaryExperiment("function"):
        method = Method(a=1).submit()

    assert method.__xpm__.job.wait() == JobState.DONE


@pytest.mark.skip()
def test_done():
    """Checks that we do not run an already done job"""
    pass


def restart_function(xp):
    restart.Restart().submit()


@pytest.mark.parametrize("terminate", restart.TERMINATES_FUNC)
def test_restart(terminate):
    """Restarting the experiment should take back running tasks"""
    restart.restart(terminate, restart_function)


def test_submitted_twice():
    """Check that a job cannot be submitted twice within the same experiment"""
    with TemporaryExperiment("duplicate", maxwait=20):
        task1 = SimpleTask(x=1).submit()
        task2 = SimpleTask(x=1).submit()
        assert task1 is task2, f"{id(task1)} != {id(task2)}"


def test_configcache():
    """Test a configuration cache"""

    with TemporaryExperiment("configcache", maxwait=20):
        task = CacheConfigTask(data=CacheConfig()).submit()

    assert task.__xpm__.job.wait() == JobState.DONE


# ---- Deprecation


class NewConfig(Config):
    __xpmid__ = "new"


@deprecate
class DeprecatedConfig(NewConfig):
    __xpmid__ = "deprecated"


class OldConfig(NewConfig):
    __xpmid__ = "deprecated"


class TaskWithDeprecated(Task):
    p: Param[NewConfig]

    def execute(self):
        pass


def checknewpaths(task_new, task_old_path):
    task_new_path = task_new.__xpm__.job.path  # type: Path

    assert task_new_path.exists(), f"New path {task_new_path} should exist"
    assert task_new_path.is_symlink(), f"New path {task_new_path} should be a symlink"

    assert task_new_path.resolve() == task_old_path


def test_tasks_deprecated_inner():
    """Test that when submitting the task, the computed identifier is the one of
    the new class"""
    with TemporaryExperiment("deprecated", maxwait=0) as xp:
        # --- Check that paths are really different first
        task_new = TaskWithDeprecated(p=NewConfig()).submit(run_mode=RunMode.DRY_RUN)
        task_old = TaskWithDeprecated(p=OldConfig()).submit(run_mode=RunMode.DRY_RUN)
        task_deprecated = TaskWithDeprecated(p=DeprecatedConfig()).submit(
            run_mode=RunMode.DRY_RUN
        )

        logging.debug("New task ID: %s", task_new.__xpm__.identifier.all.hex())
        logging.debug("Old task ID: %s", task_old.__xpm__.identifier.all.hex())
        logging.debug(
            "Old task (with deprecated flag): %s",
            task_deprecated.__xpm__.identifier.all.hex(),
        )
        assert (
            task_new.stdout() != task_old.stdout()
        ), "Old and new path should be different"

        assert (
            task_new.stdout() == task_deprecated.stdout()
        ), "Deprecated path should be the same as non deprecated"

        # --- Now check that automatic linking is performed

        # Run old task with deprecated configuration
        task_old = TaskWithDeprecated(p=OldConfig()).submit()
        task_old.wait()
        task_old_path = task_old.stdout().parent

        # Fix deprecated
        OldConfig.__xpmtype__.deprecate()
        fix_deprecated(xp.workspace.path, True, False)

        checknewpaths(task_new, task_old_path)


class NewTask(Task):
    x: Param[int]

    def execute(self):
        pass


class OldTask(NewTask):
    __xpmid__ = "deprecated"


@deprecate
class DeprecatedTask(NewTask):
    __xpmid__ = "deprecated"


def test_tasks_deprecated():
    """Test that when submitting the task, the computed identifier is the one of
    the new class"""
    with TemporaryExperiment("deprecated", maxwait=20) as xp:
        # Check that paths are really different first
        task_new = NewTask(x=1).submit(run_mode=RunMode.DRY_RUN)
        task_old = OldTask(x=1).submit(run_mode=RunMode.DRY_RUN)
        task_deprecated = DeprecatedTask(x=1).submit(run_mode=RunMode.DRY_RUN)

        assert (
            task_new.stdout() != task_old.stdout()
        ), "Old and new path should be different"
        assert (
            task_new.stdout() == task_deprecated.stdout()
        ), "Deprecated path should be the same as non deprecated"

        # OK, now check that automatic linking is performed
        task_old = OldTask(x=1).submit()
        task_old.wait()
        task_old_path = task_old.stdout().parent

        # Fix deprecated
        OldTask.__xpmtype__.deprecate()
        fix_deprecated(xp.workspace.path, True, False)

        checknewpaths(task_new, task_old_path)


class needs_java(SubmitHook):
    def __init__(self, version: int):
        self.version = version

    def spec(self):
        return self.version

    def process(self, job: Job, launcher: Launcher):
        job.environ["JAVA_HOME"] = "THE_JAVA_HOME"


@needs_java(11)
class HookedTask(Task):
    def execute(self):
        pass


def test_task_submit_hook():
    result = HookedTask().submit(run_mode=RunMode.DRY_RUN)
    assert (
        result.__xpm__.task.__xpm__.job.environ.get("JAVA_HOME", None)
        == "THE_JAVA_HOME"
    )


# --- Test for lightweight tasks


class LightweightConfig(Config):
    def __post_init__(self) -> None:
        self.data = 0


class LightweightTask(LightweightTask):
    x: Param[LightweightConfig]

    def execute(self) -> None:
        self.x.data += 1


class MyLightweightTask(Task):
    x: Param[LightweightConfig]

    def execute(self) -> None:
        assert self.x.data == 1


def test_task_lightweight():
    with TemporaryExperiment("lightweight", maxwait=20):
        x = LightweightConfig()
        lwtask = LightweightTask(x=x)
        assert (
            MyLightweightTask(x=x).add_pretasks(lwtask).submit().__xpm__.job.wait()
            == JobState.DONE
        ), "Pre-tasks should be executed"

        x_2 = LightweightConfig()
        lwtask_2 = LightweightTask(x=x)
        assert (
            MyLightweightTask(x=x_2.add_pretasks(lwtask_2))
            .add_pretasks(lwtask_2)
            .submit()
            .__xpm__.job.wait()
            == JobState.DONE
        ), "Pre-tasks should be run just once"


def test_task_lightweight_init():
    with TemporaryExperiment("lightweight_init", maxwait=20):
        x = LightweightConfig()
        lwtask = LightweightTask(x=x)
        assert (
            MyLightweightTask(x=x).submit(init_tasks=[lwtask]).__xpm__.job.wait()
            == JobState.DONE
        ), "Init tasks should be executed"



"""IPC utilities"""

from typing import Optional
from pathlib import Path
import os
import sys
import logging
from .utils import logger
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler


class IPCom:
    """IPC async thread"""

    INSTANCE: Optional["IPCom"] = None

    def __init__(self):
        self.observer = Observer()
        self.observer.start()
        self.pid = os.getpid()

    def fswatch(self, watcher: FileSystemEventHandler, path: Path, recursive=False):
        if not self.observer.is_alive():
            logging.error("Observer is not alive")

        return self.observer.schedule(
            watcher, str(path.absolute()), recursive=recursive
        )

    def fsunwatch(self, watcher):
        self.observer.unschedule(watcher)


def fork_childhandler():
    if IPCom.INSTANCE:
        logger.warning(
            "Removing IPCom instance in child process (watchers won't be copied)"
        )
        IPCom.INSTANCE = None


if sys.platform != "win32":
    os.register_at_fork(after_in_child=fork_childhandler)


def ipcom():
    # If multiprocessing, remove the IPCom instance if this does not
    # belong to our process
    if IPCom.INSTANCE is not None and IPCom.INSTANCE.pid != os.getpid():
        fork_childhandler()

    if IPCom.INSTANCE is None:
        IPCom.INSTANCE = IPCom()
        logger.info("Started IPCom instance (%s)", id(IPCom.INSTANCE))
        # IPCom.INSTANCE.start()
    return IPCom.INSTANCE



"""All classes related to localhost management
"""

import subprocess
from pathlib import Path, WindowsPath, PosixPath
import os
import threading
from experimaestro.launcherfinder import LauncherRegistry
import fasteners
import psutil

from experimaestro.locking import Lock

from . import (
    Connector,
    Process,
    ProcessBuilder,
    ProcessState,
    RedirectType,
    Redirect,
)
from experimaestro.tokens import Token, CounterToken
from experimaestro.utils import logger


class PsutilProcess(Process):
    """Wrapper for psutil process"""

    def __init__(self, pid: int):
        self._process = psutil.Process(pid)

    def wait(self) -> int:
        logger.debug("Waiting (psutil) for process with PID %s", self._process.pid)
        code = self._process.wait()
        logger.debug(
            "Finished to wait (psutil) for process with PID %s", self._process.pid
        )
        return code

    async def aio_state(self):
        if self._process.is_running():
            return ProcessState.RUNNING
        return ProcessState.FINISHED

    def __repr__(self):
        return f"PsUtil({self._process})"


class LocalProcess(Process):
    def __init__(self, process: subprocess.Popen):
        self._process = process

    def __repr__(self):
        return f"Process({self._process.pid})"

    def wait(self) -> int:
        logger.debug("Waiting (python) for process with PID %s", self._process.pid)
        code = self._process.wait()
        logger.debug(
            "Finished to wait (python) for process with PID %s", self._process.pid
        )
        return code

    async def aio_state(self):
        code = self._process.poll()
        if code is None:
            return ProcessState.RUNNING

        if code == 0:
            return ProcessState.DONE

        return ProcessState.ERROR

    def tospec(self):
        return {"type": "local", "pid": self._process.pid}

    def kill(self):
        self._process.kill()

    @staticmethod
    def fromspec(connector, spec):
        pid = spec["pid"]
        try:
            return PsutilProcess(pid)
        except psutil.NoSuchProcess:
            pass

        return None


def getstream(redirect: Redirect, write: bool):
    if redirect.type == RedirectType.FILE:
        return redirect.path.open("w" if write else "r")

    if redirect.type == RedirectType.PIPE:
        return subprocess.PIPE

    if redirect.type == RedirectType.INHERIT:
        return None

    raise NotImplementedError("For %s", redirect)


class LocalProcessBuilder(ProcessBuilder):
    def start(self):
        """Start the process"""
        stdin = getstream(self.stdin, False)
        stdout = getstream(self.stdout, True)
        stderr = getstream(self.stderr, True)

        logger.debug("Popen process")
        if self.detach:
            p = subprocess.Popen(
                self.command,
                stdin=stdin,
                stderr=stderr,
                stdout=stdout,
                env=self.environ,
                close_fds=True,
                cwd="/",
            )
        else:
            p = subprocess.Popen(
                self.command,
                stdin=stdin,
                stderr=stderr,
                stdout=stdout,
                env=self.environ,
            )

        process = LocalProcess(p)

        if self.stdout and self.stdout.type == RedirectType.PIPE:
            self.stdout.function(p.stdout)
        if self.stderr and self.stderr.type == RedirectType.PIPE:
            self.stderr.function(p.stderr)

        return process


class InterProcessLock(fasteners.InterProcessLock, Lock):
    def __init__(self, path, max_delay=-1):
        super().__init__(path)
        self.max_delay = max_delay

    def __enter__(self):
        logger.debug("Locking %s", self.path)
        if not super().acquire(blocking=True, max_delay=self.max_delay, timeout=None):
            raise threading.ThreadError("Could not acquire lock")
        logger.debug("Locked %s", self.path)
        return self

    def __exit__(self, *args):
        logger.debug("Unlocking %s", self.path)
        super().__exit__(*args)


class LocalConnector(Connector):
    INSTANCE: Connector = None

    @staticmethod
    def instance():
        if LocalConnector.INSTANCE is None:
            LocalConnector.INSTANCE = LocalConnector()
        return LocalConnector.INSTANCE

    @staticmethod
    def init_registry(registry: LauncherRegistry):
        pass

    def __init__(self, localpath: Path = None):
        localpath = localpath
        if not localpath:
            localpath = Path(
                os.environ.get("XPM_WORKDIR", "~/.local/share/experimaestro")
            ).expanduser()
        super().__init__(localpath)

    def lock(self, path: Path, max_delay: int = -1) -> Lock:
        """Returns a lockable path

        Arguments:
            path {Path} -- Path of the lockfile
            max_delay {int} -- Maximum wait duration (seconds)
        """
        return InterProcessLock(path, max_delay)

    def createtoken(self, name: str, total: int) -> Token:
        tokendir = self.localpath / "tokens"
        tokendir.mkdir(exist_ok=True, parents=True)
        return CounterToken.create(name, tokendir / ("%s.counter" % name), total)

    def processbuilder(self) -> ProcessBuilder:
        return LocalProcessBuilder()

    def resolve(self, path: Path, basepath: Path = None) -> str:
        assert isinstance(path, PosixPath) or isinstance(path, WindowsPath)
        if not basepath:
            return str(path.absolute())
        try:
            return str(path.relative_to(basepath))
        except ValueError:
            return str(path)

    def setExecutable(self, path: Path, flag: bool):
        os.chmod(path, 0o744)

    def __str__(self):
        return "LocalConnector"



import importlib
import inspect
import logging
import pkgutil
from functools import lru_cache
from importlib import import_module
from pathlib import Path
from types import ModuleType
from typing import List, Optional, Set, Tuple, Union, Dict, Type
from dataclasses import dataclass

import docutils.nodes as nodes
import sphobjinv
from docutils.frontend import OptionParser
from docutils.nodes import document
from docutils.parsers.rst import Directive, Parser, directives
from docutils.utils import new_document
from sphinx.directives.other import TocTree
from sphinx.domains.python import PyCurrentModule
from termcolor import cprint

from experimaestro import Config
from experimaestro.sphinx import PyObject


def documented_from_objects(objects_inv: Path) -> Set[str]:
    inv = sphobjinv.Inventory(objects_inv)
    return set(
        obj.name
        for obj in inv.objects
        if obj.domain == "py" and obj.role == "xpmconfig"
    )


@dataclass
class DocumentationReport:
    #: Errors while parsing
    errors: List[str]

    #: Undocumented
    undocumented: List[Config]

    #: Documented but non existing
    falsy_documented: List[Config]


def analyze(
    packages: Union[str, List[str]], documented: Set[str], skip_packages: Set[str]
) -> Tuple[bool, List[Config]]:
    """Returns the list of undocumented configurations

    :param package: package(s) to explore
    :param documented: documented symbols (fully qualified names)
    :param skip_packages: List of sub-packages that should be skipped
    :return: a tuple containing whether errors where detected when importing
        files and the list of undocumented configurations
    """

    def process(mod: ModuleType, configurations: Dict[str, Type[Config]]):
        errors = []
        for info in pkgutil.iter_modules(mod.__path__, prefix=f"{mod.__name__}."):
            try:
                logging.info("Processing %s...", info.name)
                mod = importlib.import_module(info.name)
                for x in mod.__dict__.values():
                    if inspect.isclass(x) and issubclass(x, Config):
                        configurations[f"{x.__module__}.{x.__qualname__}"] = x

            except Exception:
                logging.exception(f"Module {info.name} could not be loaded")
                errors.append(f"Module {info.name} could not be loaded")

            # Process sub-modules
            if info.ispkg:
                errors.extend(process(mod, configurations))

        return errors

    configurations = {}
    for package in list(packages):
        errors = process(import_module(package), configurations)

    # Search for undocumented
    undocumented = []
    for name, configuration in configurations.items():
        if name.startswith(f"{package}.") and name not in documented:
            if all(not name.startswith(f"{x}.") for x in skip_packages):
                undocumented.append(configuration)

    # Search for falsy documented
    falsy = [name for name in documented if name not in configurations]

    return DocumentationReport(errors, undocumented, falsy)


class autodoc(nodes.Node):
    def __init__(self, content) -> None:
        super().__init__()
        self.content = content
        self.children = []


class AutotocDirective(Directive):
    has_content = True
    optional_arguments = TocTree.optional_arguments
    option_spec = TocTree.option_spec
    required_arguments = TocTree.required_arguments

    def run(self):
        return [autodoc(self.content)]


class autoxpmconfig(nodes.Node):
    def __init__(self, name: str) -> None:
        super().__init__()
        self.name = name
        self.children = []


class AutoXPMDirective(Directive):
    has_content = PyObject.has_content
    optional_arguments = PyObject.optional_arguments
    required_arguments = PyObject.required_arguments
    option_spec = PyObject.option_spec.copy()
    option_spec.update({"members": directives.unchanged})

    def run(self):
        return [autoxpmconfig(self.arguments[0].strip())]


class currentmodule(nodes.Node):
    def __init__(self, modname: Optional[str]) -> None:
        super().__init__()
        self.modname = modname
        self.children = []


class CurrentModuleDirective(Directive):
    has_content = PyCurrentModule.has_content
    optional_arguments = PyCurrentModule.optional_arguments
    required_arguments = PyCurrentModule.required_arguments
    option_spec = PyCurrentModule.option_spec

    def run(self):
        modname = self.arguments[0].strip()
        if modname == "None":
            modname = None
        return [currentmodule(modname)]


@lru_cache
def get_parser():
    directives.register_directive("toctree", AutotocDirective)
    directives.register_directive("autoxpmconfig", AutoXPMDirective)
    directives.register_directive("currentmodule", CurrentModuleDirective)
    settings = OptionParser(components=(Parser,)).get_default_values()

    return Parser(), settings


class MyVisitor(nodes.NodeVisitor):
    def __init__(self, document: document):
        super().__init__(document)
        self.toc: Set[str] = set()
        self.currentmodule = None
        self.documented = set()

    def visit_autodoc(self, node: autodoc) -> None:
        self.toc.update(node.content)

    def visit_autoxpmconfig(self, node: autoxpmconfig) -> None:
        """Called for all other node types."""
        if node.name.find(".") == -1:
            name = f"{self.currentmodule}.{node.name}"
        else:
            name = node.name

        self.documented.add(name)
        logging.debug("[autoxpmconfig] %s / %s", node.name, name)

    def visit_currentmodule(self, node: currentmodule) -> None:
        """Called for all other node types."""
        self.currentmodule = node.modname
        logging.debug("[current module] %s", self.currentmodule)

    def unknown_visit(self, node: nodes.Node) -> None:
        """Called for all other node types."""
        pass


class DocumentVisitor:
    def __init__(self) -> None:
        self.processed = set()
        self.documented = set()
        self.errors = []

    def parse_rst(self, doc_path: Path):
        logging.info("Parsing %s", doc_path)
        input_str = doc_path.read_text()
        parser, settings = get_parser()
        document = new_document(str(doc_path.absolute()), settings)
        parser.parse(input_str, document)
        visitor = MyVisitor(document)
        document.walk(visitor)
        self.documented.update(visitor.documented)

        for to_visit in visitor.toc:
            path = (doc_path.parent / f"{to_visit}.rst").absolute()
            if path not in self.processed:
                if path.is_file():
                    self.processed.add(path)
                    self.parse_rst(path)
                else:
                    self.errors.append(f"Could not find {to_visit} in file {doc_path}")


class DocumentationAnalyzer:
    """Returns a report on the documented files"""

    def __init__(self, doc_path: Path, modules: List[str], excluded: Set[str]):
        assert doc_path.is_file()

        self.doc_path = doc_path
        self.modules = modules
        self.excluded = excluded

        # Report
        self.documentation_errors = []
        self.parsing_errors = []
        self.undocumented = []
        self.falsy_documented = []  #: Documented but not in code

    def analyze(self):
        visitor = DocumentVisitor()
        visitor.parse_rst(self.doc_path)
        self.documentation_errors = visitor.errors

        report = analyze(self.modules, visitor.documented, self.excluded)
        self.parsing_errors = report.errors
        self.undocumented = [
            f"{config.__module__}.{config.__qualname__}"
            for config in report.undocumented
        ]
        self.falsy_documented = report.falsy_documented

    def report(self):
        cprint(
            f"{len(self.documentation_errors)} errors were "
            "encountered while parsing documentation",
            "red" if self.documentation_errors else "green",
        )
        for error in self.documentation_errors:
            cprint(f"  [documentation error] {error}", "red")

        for error in self.falsy_documented:
            cprint(f"  [falsy documented] {error}", "red")

        cprint(
            f"{len(self.parsing_errors)} errors were encountered while parsing modules",
            "red" if self.parsing_errors else "green",
        )

        for error in self.parsing_errors:
            cprint(f"  [import error] {error}", "red")

        cprint(f"{len(self.undocumented)} undocumented configurations")
        for error in sorted(self.undocumented):
            cprint(f"  [undocumented] {error}", "red")

    def assert_no_undocumented(self):
        assert (
            len(self.documentation_errors) == 0
            and len(self.parsing_errors) == 0
            and len(self.undocumented) == 0
        )

    def assert_valid_documentation(self):
        """Asserts that there are no falsy documented"""
        self.assert_no_undocumented()
        assert (
            len(self.falsy_documented) == 0
        ), f"{self.falsy_documented} falsy documented members"



from experimaestro.core.context import SerializationContext
from experimaestro.core.objects import ConfigInformation
from experimaestro.utils import logger
import experimaestro.taskglobals as taskglobals
from pathlib import Path
import json


def load_job(job_path: Path, discard_id=True):
    logger.info("Loading configuration %s", job_path.parent)
    params = json.loads(job_path.resolve().read_text())
    taskglobals.Env.instance().wspath = Path(params["workspace"])

    try:
        return params, ConfigInformation.fromParameters(
            params["objects"], False, discard_id=discard_id
        )
    except Exception:
        logger.exception("Error while loading the parameters from %s", job_path)
        return None, None


def fix_deprecated(workpath: Path, fix: bool, cleanup: bool):
    jobspath = workpath / "jobs"
    logger.info("Looking for deprecated jobs in %s", jobspath)

    if cleanup:
        for job_path in jobspath.glob("*/*/params.json"):
            # If link, skip
            if job_path.parent.is_symlink():
                job_path.parent.unlink()
                logger.info("Removing symlink %s", job_path.parent)

    for job_path in jobspath.glob("*/*/params.json"):
        # If link, skip
        if job_path.parent.is_symlink():
            logger.debug("... it is a symlink - skipping")
            continue

        params, job = load_job(job_path)
        if job is None:
            continue

        # Now, computes the old  signature
        name = job_path.parents[1].name
        old_identifier: str = job_path.parent.name
        new_identifier: str = job.__xpm__.identifier.all.hex()

        if new_identifier != old_identifier:
            logger.info(
                "Configuration %s (%s) has a new identifier %s (%s)",
                name,
                old_identifier,
                job.__xpmtype__.identifier,
                new_identifier,
            )

            if fix:
                oldjobpath = jobspath / name / old_identifier
                newjobpath = jobspath / str(job.__xpmtype__.identifier) / new_identifier
                newjobpath.parent.mkdir(exist_ok=True)

                # Remove the old symlink if dandling
                if newjobpath.is_symlink() and not newjobpath.exists():
                    newjobpath.unlink()

                if newjobpath.exists():
                    if newjobpath.resolve() != oldjobpath.resolve():
                        logger.warning(
                            "New job path %s exists and is set to a "
                            "different value (%s) than the computed one (%s)",
                            newjobpath,
                            newjobpath.resolve(),
                            oldjobpath.resolve(),
                        )
                else:
                    logger.info("Fixing %s/%s", name, old_identifier)
                    if cleanup:
                        # Rewrite params.json
                        params["objects"] = job.__xpm__.__get_objects__(
                            [], SerializationContext()
                        )
                        tmppath = job_path.with_suffix(".json.tmp")
                        with tmppath.open("wt") as out:
                            json.dump(params, out)
                        tmppath.replace(job_path)

                        # Rename the folder
                        oldjobpath.rename(newjobpath)
                    else:
                        newjobpath.symlink_to(oldjobpath)



from abc import ABC, abstractmethod
import inspect
import sys
from typing import Set, Union, Dict, Iterator, List, get_args, get_origin
from collections import ChainMap
from pathlib import Path
import typing
from docstring_parser.parser import parse
import experimaestro.typingutils as typingutils
from experimaestro.utils import logger
from .arguments import Argument
from enum import Enum
import ast
import textwrap

if sys.version_info.major == 3 and sys.version_info.minor < 9:
    from typing_extensions import _AnnotatedAlias, get_type_hints
else:
    from typing import _AnnotatedAlias, get_type_hints

if typing.TYPE_CHECKING:
    from experimaestro.scheduler.base import Job
    from experimaestro.launchers import Launcher
    from experimaestro.core.objects import Config


class Identifier:
    def __init__(self, name: str):
        self.name = name

    def __hash__(self):
        return self.name.__hash__()

    def __eq__(self, other):
        return other.name.__eq__(self.name)

    def __getattr__(self, key):
        return self(key)

    def __getitem__(self, key):
        return self(key)

    def __call__(self, key):
        return Identifier(self.name + "." + key)

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


def identifier(name: Union[str, Identifier]):
    if isinstance(name, Identifier):
        return name
    return Identifier(str(name))


class Type:
    """Any experimaestro type is a child class"""

    DEFINED: Dict[type, "Type"] = {}

    def __init__(self, tn: Union[str, Identifier], description=None):
        if tn is None:
            pass
        elif isinstance(tn, str):
            tn = Identifier(tn)
        self.identifier = tn
        self._description = description

    @property
    def ignore(self):
        """Ignore by default"""
        return False

    def __str__(self):
        return "Type({})".format(self.identifier)

    def __repr__(self):
        return "Type({})".format(self.identifier)

    def name(self):
        return str(self.identifier)

    def isArray(self):
        return False

    def validate(self):
        raise NotImplementedError(f"validate ({self.__class__})")

    @staticmethod
    def fromType(key):
        """Returns the type object corresponding to the given type"""
        logger.debug("Searching for type %s", key)
        from .objects import Config

        if key is None:
            return Any

        defined = Type.DEFINED.get(key, None)
        if defined:
            return defined

        if isinstance(key, Type):
            return key

        if isinstance(key, TypeProxy):
            return key()

        if isinstance(key, Config):
            return key.__getxpmtype__()

        if inspect.isclass(key):
            if issubclass(key, Enum):
                return EnumType(key)

            if issubclass(key, Config):
                return key.__getxpmtype__()

        t = typingutils.get_list(key)
        if t:
            return ArrayType(Type.fromType(t))

        t = typingutils.get_dict(key)
        if t:
            return DictType(Type.fromType(t[0]), Type.fromType(t[1]))

        # Takes care of generics
        if get_origin(key):
            return GenericType(key)

        raise Exception("No type found for %s", key)


class DeprecatedAttribute:
    def __init__(self, fn):
        self.fn = fn
        self.key = fn.__name__
        self.warned = False

    def __set_name__(self, owner, name):
        self.key = name

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        raise NotImplementedError(f"{instance} {owner}")

    def __set__(self, instance, value):
        if not self.warned:
            logger.warning(f"Parameter {self.key} is deprecated")
        self.fn(instance, value)


class SubmitHook(ABC):
    """Hook called before the job is submitted to the scheduler

    This allows modifying e.g. the run environnement
    """

    def __call__(self, cls: typing.Type["Config"]):
        """Decorates a an XPM configuration"""
        cls.__getxpmtype__().submit_hooks.add(self)
        return cls

    @abstractmethod
    def process(self, job: "Job", launcher: "Launcher"):
        """Apply the hook for the job/launcher"""
        ...

    @abstractmethod
    def spec(self):
        """Returns an identifier tuple for hashing/equality"""
        ...

    def __eq__(self, other):
        if other.__class__ is not self.__class__:
            return False
        return self.__spec__ == other.__spec__

    def __hash__(self):
        return hash((self.__class__, self.spec()))


class XPMValue:
    """Jut marks a XPMValue"""

    pass


class ObjectType(Type):
    submit_hooks: Set[SubmitHook]
    """Hooks associated with this configuration"""

    """ObjectType contains class-level information about
    experimaestro configurations and tasks

    :param objecttype: The python Type of the associated object
    :param configtype: The python Type of the configuration object that uses
        property for arguments
    """

    # Those entries should not be copied in the __dict__
    FORBIDDEN_KEYS = set(("__dict__", "__weakref__"))

    def __init__(
        self,
        tp: type,
        identifier: Union[str, Identifier] = None,
    ):
        """Creates a type"""
        from .objects import Config, TypeConfig

        # Task related attributes
        self.taskcommandfactory = None
        self.task = None
        self._title = None
        self.submit_hooks = set()

        # Get the identifier
        if identifier is None and hasattr(tp, "__xpmid__"):
            __xpmid__ = getattr(tp, "__xpmid__")
            if isinstance(__xpmid__, Identifier):
                identifier = __xpmid__
            if inspect.ismethod(__xpmid__):
                identifier = Identifier(__xpmid__())
            elif "__xpmid__" in tp.__dict__:
                identifier = Identifier(__xpmid__)

        package = tp.__module__.lower()
        name = tp.__qualname__.lower()

        if identifier is None:
            qname = f"{package}.{name}"
            assert (
                getattr(sys, "_called_from_test", False) or "<locals>" not in qname
            ), "Configurations should not be within functions"
            identifier = Identifier(qname)

        super().__init__(identifier, None)

        # --- Creates the config type and not config type

        self.originaltype = tp
        if not issubclass(tp, Config):
            # Adds Config as a base class if not present
            __bases__ = () if tp.__bases__ == (object,) else tp.__bases__
            __dict__ = dict(tp.__dict__)

            __dict__ = {
                key: value
                for key, value in tp.__dict__.items()
                if key not in ObjectType.FORBIDDEN_KEYS
            }
            self.basetype = type(tp.__name__, (Config,) + __bases__, __dict__)
            self.basetype.__module__ = tp.__module__
            self.basetype.__qualname__ = tp.__qualname__
        else:
            self.basetype = tp

        # --- Create the type-specific configuration class (XPMConfig)
        __configbases__ = tuple(
            s.__getxpmtype__().configtype
            for s in tp.__bases__
            if issubclass(s, Config) and (s is not Config)
        ) or (TypeConfig,)

        *tp_qual, tp_name = self.basetype.__qualname__.split(".")
        self.configtype = type(
            f"{tp_name}.XPMConfig", __configbases__ + (self.basetype,), {}
        )
        self.configtype.__qualname__ = ".".join(tp_qual + [self.configtype.__name__])
        self.configtype.__module__ = tp.__module__

        # Return type is used by tasks to change the output
        if hasattr(self.basetype, "task_outputs") or False:
            self.returntype = get_type_hints(
                getattr(self.basetype, "task_outputs")
            ).get("return", typing.Any)
        else:
            self.returntype = self.basetype

        # Registers ourselves
        self.basetype.__xpmtype__ = self
        self.configtype.__xpmtype__ = self

        # Other initializations
        self.__initialized__ = False
        self._runtype = None
        self.annotations = []
        self._deprecated = False

    @property
    def objecttype(self):
        """Returns the object type"""
        return self.basetype.XPMValue

    def addAnnotation(self, annotation):
        assert not self.__initialized__
        self.annotations.append(annotation)

    def getpythontaskcommand(self, pythonpath=None):
        import experimaestro.commandline as commandline

        command = commandline.Command()
        command.add(commandline.CommandPath(pythonpath or sys.executable))
        command.add(commandline.CommandString("-m"))
        command.add(commandline.CommandString("experimaestro"))
        command.add(commandline.CommandString("run"))
        command.add(commandline.CommandParameters())
        commandLine = commandline.CommandLine()
        commandLine.add(command)

        return commandline.CommandLineTask(commandLine)

    def __initialize__(self):
        """Effectively gather configuration information"""
        # Check if not initialized
        if self.__initialized__:
            return
        self.__initialized__ = True

        from .objects import Task

        # Get the module
        module = inspect.getmodule(self.originaltype)
        if getattr(module, "__file__", None) is None:
            self._file = None
        else:
            self._file = Path(inspect.getfile(self.originaltype)).absolute()
        self._module = module.__name__
        self._package = module.__package__

        # The class of the object

        self._arguments = ChainMap(
            {}, *(tp.arguments for tp in self.parents())
        )  # type: ChainMap[Argument, Any]

        # Add arguments from annotations
        for annotation in self.annotations:
            annotation.process(self)

        # Add task
        if self.taskcommandfactory is not None:
            self.task = self.taskcommandfactory(self)
        elif issubclass(self.basetype, Task):
            self.task = self.getpythontaskcommand()

        # Add arguments from type hints
        from .arguments import TypeAnnotation

        if hasattr(self.basetype, "__annotations__"):
            typekeys = set(self.basetype.__dict__.get("__annotations__", {}).keys())
            hints = get_type_hints(self.basetype, include_extras=True)
            for key, typehint in hints.items():
                # Filter out hints from parent classes
                if key in typekeys:
                    options = None
                    if isinstance(typehint, _AnnotatedAlias):
                        for value in typehint.__metadata__:
                            if isinstance(value, TypeAnnotation):
                                options = value(options)
                        if options is not None:
                            try:
                                self.addArgument(
                                    options.create(
                                        key, self.objecttype, typehint.__args__[0]
                                    )
                                )
                            except Exception:
                                logger.error(
                                    "while adding argument %s of %s",
                                    key,
                                    self.objecttype,
                                )
                                raise

    def name(self):
        return f"{self.basetype.__module__}.{self.basetype.__qualname__}"

    def __parsedoc__(self):
        """Parse the documentation"""
        # Initialize the object if needed
        if self._title is not None:
            return
        self.__initialize__()

        # Get description from documentation
        __doc__ = self.basetype.__dict__.get("__doc__", None)
        if __doc__:
            parseddoc = parse(__doc__)
            self._title = parseddoc.short_description
            self._description = parseddoc.long_description
            for param in parseddoc.params:
                argument = self._arguments.get(param.arg_name, None)
                if argument is None:
                    logger.warning(
                        "Found documentation for undeclared argument %s", param.arg_name
                    )
                else:
                    argument.help = param.description

        # Get argument help from annotations (PEP 257)
        parsed = ast.parse(textwrap.dedent(inspect.getsource(self.originaltype)))

        argname = None  # Current argument name
        for node in parsed.body[0].body:
            if isinstance(node, ast.AnnAssign):
                argname = node.target.id
            else:
                if isinstance(node, ast.Expr) and isinstance(node.value, ast.Constant):
                    argument = self._arguments.get(argname, None)
                    if argument is not None:
                        argument.help = node.value.value

                argname = None

    def deprecate(self):
        if len(self.basetype.__bases__) != 1:
            raise RuntimeError(
                "Deprecated configurations must have "
                "only one parent (the new configuration)"
            )
        assert not self._deprecated, "Already deprecated"

        # Uses the parent identifier (and saves the deprecated one for path updates)
        self._deprecated_identifier = self.identifier
        parent = self.basetype.__bases__[0].__getxpmtype__()
        self.identifier = parent.identifier
        self._deprecated = True

    @property
    def deprecated(self) -> bool:
        """Returns true if this type is deprecated"""
        return self._deprecated

    @property
    def description(self) -> str:
        self.__parsedoc__()
        return self._description

    @property
    def title(self) -> Dict[str, Argument]:
        self.__parsedoc__()
        return self._title or str(self.identifier)

    @property
    def arguments(self) -> Dict[str, Argument]:
        self.__initialize__()
        return self._arguments

    def addArgument(self, argument: Argument):
        self._arguments[argument.name] = argument
        argument.objecttype = self

        # The the attribute for the config type
        setattr(
            self.configtype,
            argument.name,
            property(
                lambda _self: _self.__xpm__.get(argument.name),
                lambda _self, value: _self.__xpm__.set(argument.name, value),
            ),
        )

        # Check default value
        if argument.default is not None:
            argument.type.validate(argument.default)

    def getArgument(self, key: str) -> Argument:
        self.__initialize__()
        return self._arguments[key]

    def parents(self) -> Iterator["ObjectType"]:
        from .objects import Config, Task

        for tp in self.basetype.__bases__:
            if issubclass(tp, Config) and tp not in [Config, Task]:
                yield tp.__xpmtype__

    def validate(self, value):
        """Ensures that the value is compatible with this type"""
        from .objects import Config

        self.__initialize__()

        if value is None:
            return None

        if not isinstance(value, Config):
            raise ValueError(f"{value} is not an experimaestro type or task")

        types = self.basetype

        if not isinstance(value, types):
            raise ValueError(
                "%s is not a subtype of %s (MRO: %s)"
                % (value, types, ", ".join(str(cls) for cls in value.__class__.__mro__))
            )

        # Check that the task has been submitted
        if self.task and not value.__xpm__.job:
            raise ValueError("The value must be submitted before giving it")
        return value

    def fullyqualifiedname(self) -> str:
        """Returns the fully qualified (Python) name"""
        return f"{self.basetype.__module__}.{self.basetype.__qualname__}"


class TypeProxy:
    def __call__(self) -> Type:
        """Returns the real type"""
        raise NotImplementedError()


def definetype(*types):
    def call(typeclass):
        instance = typeclass(types[0].__name__)
        for t in types:
            Type.DEFINED[t] = instance
        return typeclass

    return call


@definetype(int)
class IntType(Type):
    def validate(self, value):
        if isinstance(value, float):
            import math

            rest, intvalue = math.modf(value)
            if rest != 0:
                raise TypeError(f"Value {value} is not an integer but a float")
            return int(intvalue)

        if not isinstance(value, int):
            raise TypeError(f"Value of type {type(value)} is not an integer")
        return value


@definetype(str)
class StrType(Type):
    def validate(self, value):
        if not isinstance(value, str):
            raise TypeError("value is not a string")
        return str(value)


@definetype(float)
class FloatType(Type):
    def validate(self, value):
        if not isinstance(value, (float, int)):
            raise TypeError("value is not a float")
        return float(value)


@definetype(bool)
class BoolType(Type):
    def validate(self, value):
        return bool(value)


@definetype(Path)
class PathType(Type):
    def validate(self, value):
        if isinstance(value, dict) and value.get("$type", None) == "path":
            return Path(value.get("$value"))

        if not isinstance(value, (str, Path)):
            raise TypeError("value is not a pathlike value")
        return Path(value)

    @property
    def ignore(self):
        """Ignore by default"""
        return True


class AnyType(Type):
    def __init__(self):
        super().__init__("any")

    def validate(self, value):
        return value


Any = AnyType()


class ArrayType(Type):
    def __init__(self, type: Type):
        self.type = type

    def name(self):
        return f"List[{self.type.name()}]"

    def validate(self, value):
        if not isinstance(value, List):
            raise ValueError("value is not a list")

        return [self.type.validate(x) for x in value]

    def __str__(self):
        return f"Array({self.type})"

    def __repr__(self):
        return f"Array({self.type})"


class EnumType(Type):
    def __init__(self, type: typing.Type[Enum]):
        self.type = type

    def validate(self, value):
        assert isinstance(value, self.type), f"{value} is not of type {self.type}"
        return value

    def __str__(self):
        return f"Enum({self.type})"

    def __repr__(self):
        return f"Enum({self.type})"


class DictType(Type):
    def __init__(self, keytype: Type, valuetype: Type):
        self.keytype = keytype
        self.valuetype = valuetype

    def name(self):
        return f"Dict[{self.keytype.name()},{self.valuetype.name()}]"

    def validate(self, value):
        if not isinstance(value, dict):
            raise ValueError("value is not a dict")

        return {
            self.keytype.validate(key): self.valuetype.validate(value)
            for key, value in value.items()
        }

    def __str__(self):
        return f"Dict({self.keytype.name()},{self.valuetype.name()})"

    def __repr__(self):
        return str(self)


class GenericType(Type):
    def __init__(self, type: typing.Type):
        self.type = type
        self.origin = get_origin(type)

        self.args = get_args(type)

    def name(self):
        return str(self.type)

    def __repr__(self):
        return repr(self.type)

    def validate(self, value):
        # Now, let's check generics...
        mros = typingutils.generic_mro(type(value))
        matching = next(
            (mro for mro in mros if (get_origin(mro) or mro) is self.origin), None
        )
        target = get_origin(self.type) or self.type
        if matching is None:
            raise ValueError(
                f"{type(value)} is not of type {target} ({type(value).__mro__})"
            )

        return value



from experimaestro import (
    Config,
    Task,
    Param,
    SerializationLWTask,
    copyconfig,
    state_dict,
    from_state_dict,
)
from experimaestro.core.context import SerializationContext
from experimaestro.core.objects import TypeConfig
from experimaestro.tests.utils import TemporaryExperiment


class SubModel(Config):
    pass


class Model(Config):
    submodel: Param[SubModel]

    def __post_init__(self):
        self.initialized = False
        self.submodel.initialized = False


class LoadModel(SerializationLWTask):
    def execute(self):
        self.value.initialized = True
        self.value.submodel.initialized = True


class Trainer(Task):
    model: Param[Config]

    def task_outputs(self, dep):
        model = copyconfig(self.model)
        return model.add_pretasks(dep(LoadModel(value=model)))

    def execute(self):
        assert not self.model.initialized, "Model not initialized"


class Evaluate(Task):
    model: Param[Config]
    is_submodel: Param[bool] = False

    def execute(self):
        assert self.model.initialized, "Model not initialized"
        if self.is_submodel:
            assert isinstance(self.model, SubModel)
        else:
            assert isinstance(self.model, Model)


def test_serializers_xp():
    with TemporaryExperiment("serializers", maxwait=20, port=0):
        model = Model(submodel=SubModel())
        trained_model: Model = Trainer(model=model).submit()

        # Use the model itself
        Evaluate(model=trained_model).submit()

        # Use a submodel
        Evaluate(model=trained_model.submodel, is_submodel=True).add_pretasks_from(
            trained_model
        ).submit()


class Object1(Config):
    pass


class Object2(Config):
    object: Param[Object1]


def test_serializers_serialization():
    context = SerializationContext(save_directory=None)

    obj1 = Object1()
    obj2 = Object2(object=obj1)

    data = state_dict(context, [obj1, obj2])

    [obj1, obj2] = from_state_dict(data)
    assert isinstance(obj1, Object1) and isinstance(obj1, TypeConfig)
    assert isinstance(obj2, Object2)
    assert obj2.object is obj1

    [obj1, obj2] = from_state_dict(data, as_instance=True)
    assert isinstance(obj1, Object1) and not isinstance(obj1, TypeConfig)
    assert isinstance(obj2, Object2)
    assert obj2.object is obj1



import logging
import threading
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    get_type_hints,
)
from experimaestro.connectors.local import LocalConnector
import re
from contextlib import contextmanager
from dataclasses import dataclass
from experimaestro.launcherfinder.registry import (
    LauncherRegistry,
)
from experimaestro.utils import ThreadingCondition
from experimaestro.tests.connectors.utils import OutputCaptureHandler
from experimaestro.utils.asyncio import asyncThreadcheck
from experimaestro.compat import cached_property
from experimaestro.launchers import Launcher
from experimaestro.scriptbuilder import PythonScriptBuilder
from experimaestro.connectors import (
    Connector,
    ProcessBuilder,
    Process,
    ProcessState,
    Redirect,
    RedirectType,
)

logger = logging.getLogger("xpm.slurm")


class SlurmJobState:
    start: str
    end: str
    status: str

    STATE_MAP = {
        "CONFIGURING": ProcessState.SCHEDULED,
        "REQUEUE_FED": ProcessState.SCHEDULED,
        "PENDING": ProcessState.SCHEDULED,
        "REQUEUE_HOLD": ProcessState.SCHEDULED,
        "COMPLETING": ProcessState.RUNNING,
        "RUNNING": ProcessState.RUNNING,
        "COMPLETED": ProcessState.DONE,
        "FAILED": ProcessState.ERROR,
        "DEADLINE": ProcessState.ERROR,
        "NODE_FAIL": ProcessState.ERROR,
        "REVOKED": ProcessState.ERROR,
        "TIMEOUT": ProcessState.ERROR,
        "CANCELLED": ProcessState.ERROR,
        "BOOT_FAIL": ProcessState.ERROR,
        "OUT_OF_MEMORY": ProcessState.ERROR,
    }

    def __init__(self, status, start, end):
        self.slurm_state = status if status[-1] == "+" else status
        if self.slurm_state.startswith("CANCELLED"):
            self.state = ProcessState.ERROR
        else:
            self.state = SlurmJobState.STATE_MAP.get(self.slurm_state, None)
            if self.state is None:
                logging.warning(
                    "Unknown state: %s (supposing this is an error)", self.slurm_state
                )
                self.state = ProcessState.ERROR

        self.start = start
        self.end = end

    def finished(self):
        """Returns true if the job has finished"""
        return self.state.finished

    def __repr__(self):
        return f"{self.slurm_state} ({self.start}-{self.end})"


class SlurmProcessWatcher(threading.Thread):
    """Process that calls sacct at regular interval to check job status"""

    WATCHERS: Dict[Tuple[Tuple[str, Any]], "SlurmProcessWatcher"] = {}

    def __init__(self, launcher: "SlurmLauncher"):
        super().__init__()
        self.launcher = launcher
        self.count = 1
        self.jobs: Dict[str, SlurmJobState] = {}

        self.cv = ThreadingCondition()
        self.fetched_event = threading.Event()
        self.updating_jobs = threading.Lock()
        self.start()

    @staticmethod
    @contextmanager
    def get(launcher: "SlurmLauncher"):
        watcher = SlurmProcessWatcher.WATCHERS.get(launcher.key, None)
        if watcher is None:
            watcher = SlurmProcessWatcher(launcher)
            SlurmProcessWatcher.WATCHERS[launcher.key] = watcher
        else:
            with watcher.cv:
                watcher.count += 1
        yield watcher
        watcher.count -= 1
        with watcher.cv:
            watcher.cv.notify()

    def getjob(self, jobid, timeout=None):
        """Allows to share the calls to sacct"""

        # Ensures that we have fetched at least once
        self.fetched_event.wait()

        # Waits that jobs are refreshed (with a timeout)
        with self.cv:
            self.cv.wait(timeout=timeout)

        # Ensures jobs are not updated right now
        with self.updating_jobs:
            return self.jobs.get(jobid)

    def run(self):
        while self.count > 0:
            builder = self.launcher.connector.processbuilder()
            builder.command = [
                f"{self.launcher.binpath}/sacct",
                "-n",
                "-p",
                "--format=JobID,State,Start,End",
            ]
            handler = OutputCaptureHandler()
            builder.detach = False
            builder.stdout = Redirect.pipe(handler)
            builder.environ = self.launcher.launcherenv
            logger.debug("Checking SLURM state with sacct")
            process = builder.start()

            with self.updating_jobs:
                self.jobs = {}
                output = handler.output.decode("utf-8")
                for line in output.split("\n"):
                    line = line.strip()
                    if line:
                        try:
                            jobid, state, start, end, *_ = line.split("|")
                            self.jobs[jobid] = SlurmJobState(state, start, end)
                            logger.debug("Parsed line: %s", line)
                        except ValueError:
                            logger.error("Could not parse line %s", line)
            process.kill()

            with self.cv:
                logger.debug("Jobs %s", self.jobs)
                self.fetched_event.set()
                self.cv.notify_all()

                self.cv.wait_for(
                    lambda: self.count == 0, timeout=self.launcher.interval
                )
                if self.count == 0:
                    logger.debug("Stopping SLURM watcher process")
                    del SlurmProcessWatcher.WATCHERS[self.launcher.key]
                    break


class BatchSlurmProcess(Process):
    """A batch slurm process"""

    def __init__(self, launcher: "SlurmLauncher", jobid: str):
        self.launcher = launcher
        self.jobid = jobid

    def wait(self):
        with SlurmProcessWatcher.get(self.launcher) as watcher:
            while True:
                state = watcher.getjob(self.jobid)
                if state and state.finished():
                    return 0 if state.slurm_state == "COMPLETED" else 1

    async def aio_state(self):
        def check():
            with SlurmProcessWatcher.get(self.launcher) as watcher:
                jobinfo = watcher.getjob(self.jobid)
                return jobinfo.state if jobinfo else ProcessState.SCHEDULED

        return await asyncThreadcheck("slurm.aio_isrunning", check)

    def kill(self):
        logger.warning("Killing slurm job %s", self.jobid)
        builder = self.launcher.connector.processbuilder()
        builder.command = [f"{self.launcher.binpath}/scancel", f"{self.jobid}"]
        builder.start()

    def __repr__(self):
        return f"slurm:{self.jobid}"

    def tospec(self):
        return {"type": "slurm", "pid": self.jobid, "options": self.launcher.key}

    @classmethod
    def fromspec(cls, connector: Connector, spec: Dict[str, Any]):
        options = {k: v for k, v in spec.get("options", ())}
        launcher = SlurmLauncher(connector=connector, **options)
        process = BatchSlurmProcess(launcher, spec["pid"])

        # Checks that the process is running
        with SlurmProcessWatcher.get(launcher) as watcher:
            logger.info("Checking SLURM job %s", process.jobid)
            jobinfo = watcher.getjob(process.jobid, timeout=0.1)
            if jobinfo and jobinfo.state.running:
                logger.debug(
                    "SLURM job is running (%s), returning process", process.jobid
                )
                return process
        return None


def addstream(command: List[str], option: str, redirect: Redirect):
    if redirect.type == RedirectType.FILE:
        command.extend([option, redirect.path])
    elif redirect.type == RedirectType.INHERIT:
        pass
    else:
        raise NotImplementedError("For %s", redirect)


class SlurmProcessBuilder(ProcessBuilder):
    def __init__(self, launcher: "SlurmLauncher"):
        super().__init__()
        self.launcher = launcher

    def start(self) -> BatchSlurmProcess:
        """Start the process"""
        builder = self.launcher.connector.processbuilder()
        builder.workingDirectory = self.workingDirectory
        builder.environ = self.launcher.launcherenv
        builder.detach = False

        if not self.detach:
            # Simplest case: we wait for the output
            builder.command = [f"{self.launcher.binpath}/srun"]
            builder.command.extend(self.launcher.options.args())
            builder.command.extend(self.command)
            builder.stdin = self.stdin
            builder.stdout = self.stdout
            builder.stderr = self.stderr
            builder.environ = self.environ
            builder.detach = False
            return builder.start()

        builder.command = [f"{self.launcher.binpath}/sbatch", "--parsable"]
        builder.command.extend(self.launcher.options.args())

        addstream(builder.command, "-e", self.stderr)
        addstream(builder.command, "-o", self.stdout)
        addstream(builder.command, "-i", self.stdin)

        builder.command.extend(self.command)
        logger.info("slurm sbatch command: %s", builder.command)
        handler = OutputCaptureHandler()
        builder.stdout = Redirect.pipe(handler)
        builder.stderr = Redirect.inherit()
        p = builder.start()
        if p.wait() != 0:
            logger.error("Error while running sbatch command")
            raise RuntimeError("Error while submitting job")

        output = handler.output.decode("utf-8").strip(" \n")
        RE_SUBMITTED_JOB = re.compile(r"""^(\d+)(?:;.*)?$""", re.MULTILINE)
        m = RE_SUBMITTED_JOB.match(output)
        if m is None:
            raise RuntimeError(f"Could not get the submitted job ID from {output}")

        return BatchSlurmProcess(self.launcher, m.group(1))


@dataclass
class SlurmOptions:
    # Options
    nodes: Optional[int] = 1
    """Number of requested nodes"""

    time: Optional[str] = None
    """Requested time"""

    account: Optional[str] = None
    """The account for launching the job"""

    qos: Optional[str] = None
    """The requested Quality of Service"""

    partition: Optional[str] = None
    """The requested partition"""

    constraint: Optional[str] = None
    """Logic expression on node features (as defined by the administator)"""

    mem: Optional[str] = None
    """Requested memory on the node (in megabytes by default)"""

    exclude: Optional[str] = None
    """List of hosts to exclude"""

    mem_per_gpu: Optional[str] = None
    """Requested memory per allocated GPU (size with units: K, M, G, or T)"""

    cpus_per_task: Optional[str] = None
    """Number of cpus requested per task"""

    nodelist: Optional[str] = None
    """Request a specific list of hosts"""

    # GPU-related
    gpus: Optional[int] = None
    """Number of GPUs"""

    gpus_per_node: Optional[int] = None
    """Number of GPUs per node"""

    def args(self) -> List[str]:
        """Returns the corresponding options"""
        options = []
        for key in get_type_hints(SlurmOptions).keys():
            value = getattr(self, key, None)
            if value is not None:
                options.append(
                    f"""--{key.replace("_", "-")}={value}""",
                )

        return options

    def merge(self, other: "SlurmOptions"):
        merged = SlurmOptions()

        for key, value in self.__dict__.items():
            setattr(merged, key, value)

        for key, value in other.__dict__.items():
            if value is not None:
                setattr(merged, key, value)

        return merged

    @staticmethod
    def format_time(duration_s: int):
        """Format time for the SLURM option

        :param duration_s: Time duration in seconds1
        :return: The configuration string
        """
        seconds = duration_s % 60
        minutes = (duration_s // 60) % 60
        hours = duration_s // 3600
        return f"{hours}:{minutes}:{seconds}"


class SlurmLauncher(Launcher):
    """Slurm workload manager launcher

    https://slurm.schedmd.com/documentation.html
    """

    def __init__(
        self,
        *,
        connector: Connector = None,
        options: SlurmOptions = None,
        interval: float = 60,
        main=None,
        launcherenv: Dict[str, str] = None,
        binpath="/usr/bin",
    ):
        """
        Arguments:
            main: Main slurm launcher to avoid launching too many polling jobs
            interval: seconds between polling job statuses
        """
        super().__init__(connector or LocalConnector.instance())
        self.binpath = Path(binpath)
        self.interval = interval
        self.launcherenv = launcherenv
        self.options = options or SlurmOptions()

    def __str__(self):
        return f"SlurmLauncher({self.options}, path={self.binpath})"

    @staticmethod
    def init_registry(registry: LauncherRegistry):
        from .configuration import SlurmConfiguration

        registry.register_launcher("slurm", SlurmConfiguration)

    @staticmethod
    def get_cli():
        from .cli import cli

        return cli

    @cached_property
    def key(self):
        """Returns a dictionary characterizing this launcher when calling sacct/etc"""
        return (
            ("binpath", str(self.binpath)),
            ("interval", self.interval),
        )

    def config(self, **kwargs):
        """Returns a new Slurm launcher with the given configuration"""
        return SlurmLauncher(
            connector=self.connector,
            binpath=self.binpath,
            launcherenv=self.launcherenv,
            options=self.options.merge(SlurmOptions(**kwargs)),
            interval=self.interval,
        )

    def scriptbuilder(self):
        """Returns the script builder

        We assume *nix, but should be changed to PythonScriptBuilder when working
        """
        builder = PythonScriptBuilder()
        builder.processtype = "slurm"
        return builder

    def processbuilder(self) -> SlurmProcessBuilder:
        """Returns the process builder for this launcher

        By default, returns the associated connector builder"""
        return SlurmProcessBuilder(self)



from pathlib import Path
import sys
from experimaestro.connectors import Redirect
from experimaestro.connectors.local import LocalProcessBuilder
from experimaestro.tests.connectors.utils import OutputCaptureHandler
from experimaestro.tests.utils import TemporaryDirectory


def test_local_simpole():
    builder = LocalProcessBuilder()

    with TemporaryDirectory() as tmp:
        builder.command = [
            sys.executable,
            Path(__file__).parent / "bin" / "executable.py",
        ]
        builder.workingDirectory = tmp
        outpath = tmp / "stdout.txt"
        builder.stdout = Redirect.file(outpath)

        p = builder.start()
        p.wait()

        assert outpath.read_text().strip() == "hello world"


def test_local_pipe():
    builder = LocalProcessBuilder()

    with TemporaryDirectory() as tmp:
        builder.command = [
            sys.executable,
            Path(__file__).parent / "bin" / "executable.py",
        ]
        builder.workingDirectory = tmp

        handler = OutputCaptureHandler()
        builder.stdout = Redirect.pipe(handler)

        p = builder.start()
        assert p.wait() == 0

        assert handler.output.strip() == b"hello world"



import time
from typing import List
from experimaestro import (
    param,
    Param,
    Task,
    Config,
    pathoption,
    STDOUT,
    cache,
)


class SimpleTask(Task):
    x: Param[int]

    def execute(self):
        print(self.x)  # noqa: T201


@pathoption("out", STDOUT)
class Say(Task):
    word: Param[str]

    def execute(self):
        self.out.write_text(self.word.upper())


class Concat(Task):
    strings: Param[List[Say]]

    def execute(self):
        # We access the file where standard output was stored
        says = []
        for string in self.strings:
            with open(string.out) as fp:
                says.append(fp.read().strip())
        print(" ".join(says))  # noqa: T201


@param("x", type=int)
class ForeignClassB1(Config):
    pass


@param("b", type=ForeignClassB1)
class ForeignTaskA(Task):
    def execute(self):
        print(self.b.x)  # noqa: T201


@pathoption("wait", "wait")
class Fail(Task):
    def execute(self):
        while not self.wait.is_file():
            time.sleep(0.01)
        raise AssertionError("Failing")

    def touch(self):
        while self.__xpm__.job.state.notstarted():
            time.sleep(0.01)

        with open(self.wait, "w") as out:
            out.write("hello")


@param("fail", Fail)
class FailConsumer(Task):
    def execute(self):
        return True


@param("a", int)
class Method(Task):
    def execute(self):
        assert self.a == 1


class SetUnknown(Task):
    def execute(self):
        self.abc = 1


"""Check that config works properly"""


class CacheConfig(Config):
    @cache("cached")
    def get(self, path):
        if not path.is_file():
            path.write_text("hello")
        return path.read_text()


@param("data", type=CacheConfig)
class CacheConfigTask(Task):
    def execute(self):
        assert self.data.get() == "hello"



import sys
import typing
from typing import Generic, Protocol

if sys.version_info.major == 3:
    if sys.version_info.minor < 11:
        from typing import _collect_type_vars as _collect_parameters
    else:
        from typing import _collect_parameters

    if sys.version_info.minor < 9:
        from typing_extensions import (
            _AnnotatedAlias as AnnotatedAlias,
            get_args,
            get_origin,
        )
    else:
        from typing import _AnnotatedAlias as AnnotatedAlias, get_args, get_origin

GenericAlias = typing._GenericAlias


def isgenericalias(typehint):
    """Returns True if it is a generic type alias"""
    # Works with Python 3.7, 3.8 and 3.9
    return isinstance(typehint, GenericAlias)


def get_optional(typehint):
    if isgenericalias(typehint) and typehint.__origin__ == typing.Union:
        if len(typehint.__args__) == 2:
            for ix in (0, 1):
                argtype = typehint.__args__[ix]
                origin = get_origin(argtype) or argtype
                if issubclass(origin, type(None)):
                    return typehint.__args__[1 - ix]
    return None


def get_list_component(typehint):
    """Returns the list component (or None if not a list)"""
    if isgenericalias(typehint):
        # Python 3.6, 3.7+ test
        if typehint.__origin__ in [typing.List, list]:
            assert len(typehint.__args__) == 1
            return typehint.__args__[0]
    return None


get_list = get_list_component
"""
    :deprecated: use get_list_component
"""


def is_annotated(typehint):
    return isinstance(typehint, AnnotatedAlias)


def get_type(typehint):
    """Returns the type discarding Annotated and optional"""
    while True:
        if t := get_optional(typehint):
            typehint = t
        if isinstance(typehint, AnnotatedAlias):
            typehint = get_args(typehint)[0]
        else:
            break
    return typehint


def get_dict(typehint):
    if isgenericalias(typehint):
        # Python 3.6, 3.7+ test
        if typehint.__origin__ in [typing.Dict, dict]:
            assert len(typehint.__args__) == 2
            return typehint.__args__
    return None


# From https://github.com/python/typing/issues/777


def _generic_mro(result, tp):
    origin = typing.get_origin(tp) or tp

    result[origin] = tp
    if hasattr(origin, "__orig_bases__"):
        parameters = _collect_parameters(origin.__orig_bases__)
        substitution = dict(zip(parameters, get_args(tp)))
        for base in origin.__orig_bases__:
            if get_origin(base) in result:
                continue
            base_parameters = getattr(base, "__parameters__", ())
            if base_parameters:
                base = base[tuple(substitution.get(p, p) for p in base_parameters)]
            _generic_mro(result, base)


def generic_mro(tp):
    origin = get_origin(tp)
    if origin is None and not hasattr(tp, "__orig_bases__"):
        if not isinstance(tp, type):
            raise TypeError(f"{tp!r} is not a type or a generic alias")
        return tp.__mro__
    # sentinel value to avoid to subscript Generic and Protocol
    result = {Generic: Generic, Protocol: Protocol}
    _generic_mro(result, tp)
    cls = origin if origin is not None else tp
    return tuple(result.get(sub_cls, sub_cls) for sub_cls in cls.__mro__)



from pathlib import Path, PosixPath
from typing import Callable, Dict, List, Optional
from experimaestro.commandline import AbstractCommand, Job, CommandLineJob
from experimaestro.connectors import Connector
from experimaestro.connectors.local import ProcessBuilder, LocalConnector
from experimaestro.connectors.ssh import SshPath, SshConnector
from abc import ABC, abstractmethod


class ScriptBuilder(ABC):
    """A script builder is responsible for generating the script
    used to launch a command line job"""

    lockfiles: List[Path]
    """The files that must be locked before starting the job"""

    command: "AbstractCommand"
    """Command to be run"""

    @abstractmethod
    def write(self, job: CommandLineJob) -> Path:
        """Write the commmand line job

        :params job: The job to be written
        """
        ...


SubmitListener = Callable[[Job], None]
"""Listen to job submissions"""


class Launcher(ABC):
    """A launcher"""

    submit_listeners: List[SubmitListener]

    def __init__(self, connector: Connector):
        self.connector = connector
        self.environ: Dict[str, str] = {}
        self.notificationURL: Optional[str] = None
        self.submit_listeners = []

    def setenv(self, key: str, value: str):
        self.environ[key] = value

    def setNotificationURL(self, url: Optional[str]):
        self.notificationURL = url

    @abstractmethod
    def scriptbuilder(self) -> ScriptBuilder:
        """Returns a script builder"""
        ...

    def addListener(self, listener: SubmitListener):
        self.submit_listeners.append(listener)

    def onSubmit(self, job: Job):
        """Called when submitting a job

        Example of use: this allows the launcher to add token dependencies
        """
        for listener in self.submit_listeners:
            listener(job)

    def processbuilder(self) -> ProcessBuilder:
        """Returns the process builder for this launcher

        By default, returns the associated connector builder"""
        return self.connector.processbuilder()

    @staticmethod
    def get(path: Path):
        """Get a default launcher for a given path"""
        if isinstance(path, PosixPath):
            from .direct import DirectLauncher

            return DirectLauncher(LocalConnector())

        if isinstance(path, SshPath):
            from .direct import DirectLauncher

            return DirectLauncher(SshConnector.fromPath(path))
        raise ValueError("Cannot create a default launcher for %s", type(path))



import click

"""Defines the task command line argument prefix for experimaestro-handled command lines"""


@click.group()
def cli():
    """Main entry point for CLI"""
    pass


class forwardoptionMetaclass(type):
    def __getattr__(self, key):
        """Access to a class field"""
        return forwardoption([key])


class forwardoption(metaclass=forwardoptionMetaclass):
    """Allows to access an argument of the configuration

    This allows to refer to a path of a class in a "python" syntax, e.g.
    `@forwardoption.ranker.optimizer.epsilon(MyConfig)` or
    `@forwardoption.ranker.optimizer.epsilon(MyConfig, "option-name")`

    default can be changed by setting the option
    """

    def __init__(self, path=[]):
        self.path = path

    def __call__(self, cls, option_name=None, **kwargs):
        """ """
        argument = cls.__getxpmtype__().arguments[self.path[0]]
        for c in self.path[1:]:
            argument = getattr(argument, c)

        name = "--%s" % (option_name or argument.name.replace("_", "-"))
        default = kwargs["default"] if "default" in kwargs else argument.default

        # TODO: set the type of the option when not a simple type
        return click.option(name, help=argument.help or "", default=default)

    def __getattr__(self, key):
        """Access to a class field"""
        return forwardoption([key])



"""Tokens are special types of dependency controlling the access to 
a computational resource (e.g. number of launched jobs, etc.)
"""

from dataclasses import dataclass
import sys
from pathlib import Path
from experimaestro.core.objects import Config
import fasteners
import threading
import os.path
from watchdog.events import FileSystemEventHandler
from typing import Dict
from experimaestro.launcherfinder.base import TokenConfiguration

from experimaestro.launcherfinder.registry import LauncherRegistry
from experimaestro.utils.yaml import YAMLDict

from .ipc import ipcom
from .locking import Lock, LockError
from .scheduler.dependencies import Dependency, DependencyStatus, Resource
import logging
import json


logger = logging.getLogger("xpm.tokens")


class Token(Resource):
    """Base class for all token-based resources"""

    available: int

    def aio_notify(self):
        # Notifying
        def check(dependency: Dependency):
            if self.available > 0:
                dependency.check()

        with self.dependents as dependents:
            for _dependency in dependents:
                _dependency.loop.call_soon_threadsafe(check, _dependency)


class CounterTokenLock(Lock):
    def __init__(self, dependency: "CounterTokenDependency"):
        super().__init__()
        self.dependency = dependency

    def _acquire(self):
        self.dependency.token.acquire(self.dependency)

    def _release(self):
        self.dependency.token.release(self.dependency)

    def __str__(self):
        return "Lock(%s)" % self.dependency


class CounterTokenDependency(Dependency):
    def __init__(self, token: "CounterToken", count: int):
        super().__init__(token)
        self._token = token
        self.count = count

    @property
    def name(self):
        return f"{self.target.identifier}.token"

    def status(self) -> DependencyStatus:
        if self.count <= self.token.available:
            return DependencyStatus.OK
        return DependencyStatus.WAIT

    def lock(self) -> "Lock":
        return CounterTokenLock(self)

    @property
    def token(self):
        return self._token


class TokenFile:
    """Represents a token file"""

    def __init__(self, path: Path):
        try:
            self.path = path
            with path.open("rt") as fp:
                count, self.uri = [l.strip() for l in fp.readlines()]
                self.count = int(count)
        except Exception:
            logging.exception("Error while reading %s", self.path)
            raise

    @staticmethod
    def create(dependency: CounterTokenDependency):
        path = dependency._token.path / dependency.name
        count = dependency.count
        uri = str(dependency.target.basepath)

        self = object.__new__(TokenFile)
        self.count = count
        self.uri = uri
        self.path = path
        logging.debug("Writing token file %s", path)
        with path.open("wt") as fp:
            fp.write(f"{str(count)}\n{uri}\n")
        return self

    def delete(self):
        if self.path.is_file():
            logging.debug("Deleting token file %s", self.path)
            self.path.unlink()

    def watch(self):
        """Watch the matching process"""
        logger.debug(
            "Watching process for %s (%s, taken %d)", self.path, self.uri, self.count
        )
        path = Path(self.uri)
        lockpath = path.with_suffix(".lock")
        pidpath = path.with_suffix(".pid")

        # Watch for the job
        def run():
            logger.debug("Locking job lock path %s", lockpath)
            process = None
            with fasteners.InterProcessLock(lockpath):
                if not pidpath.is_file():
                    logger.debug("Job already finished (no PID file)")
                else:
                    s = ""
                    while s == "":
                        s = pidpath.read_text()

                    logger.info("Loading job watcher from definition")
                    from experimaestro.connectors import Process

                    # FIXME: not always localhost...
                    from experimaestro.connectors.local import LocalConnector

                    connector = LocalConnector.instance()
                    process = Process.fromDefinition(connector, json.loads(s))

            # Wait out of the lock
            if process is not None:
                # Process is None: process has finished
                process.wait()

            self.delete()

        threading.Thread(target=run).start()


class CounterToken(Token, FileSystemEventHandler):
    """File-based counter token

    To ensure recovery (server stopped for whatever reason), we use one folder
    per token; inside this folder:

    - token.lock is used for IPC locking
    - token.info contains the maximum number of tokens
    - TIMESTAMP.token contains (1) the number of tokens (2) the job URI
    """

    """Maps paths to instances"""
    TOKENS: Dict[str, "CounterToken"] = {}

    @staticmethod
    def forkhandler():
        CounterToken.TOKENS = {}

    @staticmethod
    def create(name: str, path: Path, count: int):
        created = CounterToken.TOKENS.get(name, None)
        if created:
            logger.warning("Re-using token for path %s", path)
        else:
            created = CounterToken(name, path, count)
            CounterToken.TOKENS[name] = created
        return created

    @staticmethod
    def init_registry(registry: LauncherRegistry):
        registry.register_token("countertoken", CounterTokenConfiguration)

    def __init__(self, name: str, path: Path, count: int, force=True):
        """[summary]

        Arguments:
            path {Path} -- The file path of the token file
            count {int} -- Number of tokens (overrides previous definitions)
            force --   If the token has already been created, force to write the maximum
                       number of tokens
        """
        super().__init__()

        self.path = path
        self.path.mkdir(exist_ok=True, parents=True)

        self.cache: Dict[str, TokenFile] = {}

        self.infopath = path / "token.info"

        self.ipc_lock = fasteners.InterProcessLock(path / "token.lock")
        self.lock = threading.Lock()

        self.name = name

        # Set the new number of tokens
        with self.lock, self.ipc_lock:
            # Get the maximum number of tokens
            if force or not self.infopath.is_file():
                self.total = count
                self.infopath.write_text(str(count))

            self.timestamp = os.path.getmtime(self.path)
            self._update()

        # Watched path
        self.watchedpath = str(path.absolute())
        self.watcher = ipcom().fswatch(self, self.path, recursive=True)
        logger.info("Watching %s", self.watchedpath)

    def _update(self):
        """Update the state by reading all the information from disk

        Assumes that the IPC lock is taken
        """
        logging.debug("Full token state update")
        self.total = int(self.infopath.read_text())
        old_cache = self.cache
        self.cache = {}
        self.available = self.total

        for path in self.path.glob("*.token"):
            tf = old_cache.get(path.name)
            if tf is None:
                tf = TokenFile(path)
                tf.watch()
                logging.debug("Read token file %s (%d)", path, tf.count)
            else:
                logging.debug(
                    "Token file already in cache %s (%d)", path.name, tf.count
                )

            self.cache[path.name] = tf
            self.available -= tf.count
        logging.debug("Full token state update finished (%d available)", self.available)

    def __str__(self):
        return "token[{}]".format(self.name)

    def on_deleted(self, event):
        logger.debug(
            "Deleted path notification %s [watched %s]",
            event.src_path,
            self.watchedpath,
        )
        name = Path(event.src_path).name
        # Name is in cache if we did not release the token ourselves
        if name in self.cache:
            with self.lock:
                if name in self.cache:
                    logging.debug("Deleting %s from token cache (event)", name)
                    fc = self.cache[name]
                    del self.cache[name]

                    self.available += fc.count
                    logger.debug(
                        "Getting back %d tokens (%d available)",
                        fc.count,
                        self.available,
                    )

            # Do not lock here (notify only)
            if self.available > 0:
                self.aio_notify()

    def on_created(self, event):
        logger.debug(
            "Created path notification %s [watched %s]",
            event.src_path,
            self.watchedpath,
        )

        path = Path(event.src_path)

        try:
            if path.name.endswith(".token") and path.name not in self.cache:
                with self.lock:
                    if path.name not in self.cache:
                        tokenfile = TokenFile(path)
                        tokenfile.watch()
                        self.cache[path.name] = tokenfile
        except Exception:
            logger.exception("Uncaught exception in on_modified handler")
            raise

    def on_modified(self, event):
        try:
            logger.debug(
                "Watched path notification %s [watched %s]",
                event.src_path,
                self.watchedpath,
            )
            # logger.debug("%s", event)

            path = Path(event.src_path)

            if event.src_path == str(self.infopath):
                logger.debug("Token information modified")
                timestamp = os.path.getmtime(self.infopath)
                if timestamp <= self.timestamp:
                    logger.debug(
                        "Not reading token file [%f <= %f]", timestamp, self.timestamp
                    )

                total = int(self.infopath.read_text())
                delta = total - self.total
                self.total = total
                self.available += delta
                logger.debug(
                    "Token information modified: available %d, total %d",
                    self.available,
                    self.total,
                )

                if delta > 0 and self.available > 0:
                    with self.dependents as dependents:
                        for dependency in dependents:
                            dependency.check()

            elif path.name.endswith(".token") and path.name not in self.cache:
                with self.lock:
                    if path.name not in self.cache:
                        logger.debug("Token file not in cache %s", path.name)
                        tokenfile = TokenFile(path)
                        tokenfile.watch()
                        self.cache[path.name] = tokenfile
        except Exception:
            logger.exception("Uncaught exception in on_modified handler")
            raise

    def dependency(self, count):
        """Create a token dependency"""
        return CounterTokenDependency(self, count)

    def __call__(self, count, task: Config):
        """Create a token dependency and add it to the task"""
        return task.add_dependencies(self.dependency(count))

    def acquire(self, dependency: CounterTokenDependency):
        """Acquire requested token"""
        with self.lock, self.ipc_lock:
            self._update()
            if self.available < dependency.count:
                logger.warning(
                    "Not enough available (%d available, %d requested)",
                    self.available,
                    dependency.count,
                )
                raise LockError("No token")

            self.available -= dependency.count

            self.cache[dependency.name] = TokenFile.create(dependency)
            logger.debug(
                "Token state [acquired %d]: available %d, taken %d",
                dependency.count,
                self.available,
                self.total,
            )

    def release(self, dependency: CounterTokenDependency):
        """Release"""
        with self.lock, self.ipc_lock:
            self._update()

            tf = self.cache.get(dependency.name, None)
            if tf is None:
                logging.error(
                    "Could not find the taken token for %s (%s)",
                    dependency,
                    dependency.name,
                )
                return

            logging.debug("Deleting %s from token cache", dependency.name)
            del self.cache[dependency.name]
            self.available += tf.count
            logging.debug("%s: available %d", self, self.available)
            tf.delete()

        self.aio_notify()


class ProcessCounterToken(Token):
    """Process-level token"""

    def __init__(self, count: int):
        """Creates a new

        Arguments:
            count {int} -- Number of tokens
        """
        super().__init__()

        self.count = count
        self.available = count
        self.lock = threading.Lock()

    def __str__(self):
        return "process-token()"

    def dependency(self, count):
        """Create a token dependency"""
        return CounterTokenDependency(self, count)

    def __call__(self, count, task: Config):
        """Create a token dependency and add it to the task"""
        return task.add_dependencies(self.dependency(count))

    def acquire(self, dependency: CounterTokenDependency):
        """Acquire requested token"""
        with self.lock:
            if self.available < dependency.count:
                raise LockError("No token")

            self.available -= dependency.count

    def release(self, dependency: CounterTokenDependency):
        """Release"""
        with self.lock:
            self.available += dependency.count
            logging.debug(
                "%s: releasing %d (available %d)",
                self,
                dependency.count,
                self.available,
            )

        self.aio_notify()


if sys.platform != "win32":
    os.register_at_fork(after_in_child=CounterToken.forkhandler)


@dataclass
class CounterConfiguration(TokenConfiguration):
    tokens: int

    def create(self, registry: "LauncherRegistry", identifier: str):
        from experimaestro.connectors.local import LocalConnector

        return LocalConnector.instance().createtoken(identifier, self.tokens)


class CounterTokenConfiguration(YAMLDict[CounterConfiguration]):
    pass



import codecs
from collections import defaultdict
from copy import deepcopy
import io
import math
from attr import Factory
from attrs import define
import logging
from experimaestro import Annotated
from typing import (
    Dict,
    List,
    Optional,
    Set,
    TextIO,
)
import re
import humanfriendly
from dataclasses import dataclass, field
from experimaestro.launcherfinder import YAMLDataClass, HostRequirement
from experimaestro.launcherfinder.base import LauncherConfiguration
from experimaestro.launcherfinder.registry import (
    Initialize,
    LauncherRegistry,
)
from experimaestro.launcherfinder.specs import (
    CPUSpecification,
    CudaSpecification,
    HostSpecification,
)
from experimaestro.compat import cached_property
from . import Launcher
from experimaestro.connectors import (
    Redirect,
)

logger = logging.getLogger("xpm.slurm")


def fill_nodes_configuration(input: TextIO, configuration: "SlurmConfiguration"):
    """Parses the output of scontrol show nodes"""
    re_nodename = re.compile(r"""^NodeName=([_\-\w]+)""")
    re_features = re.compile(r"""^\s*AvailableFeatures=([,_\-\w]+)""")
    re_partitions = re.compile(r"""^\s*Partitions=([,\-_\w]+)""")

    nodename = ""
    features = []
    partition_names = []
    partitions = configuration.partitions
    partitions2features2nodes = defaultdict(lambda: {})

    def process():
        for partition_name in partition_names:
            partition = partitions.setdefault(partition_name, SlurmPartition(nodes=[]))

            fl = "&".join(sorted(features))
            nodes = partitions2features2nodes[partition_name].get(fl)
            if nodes is None:
                nodes = SlurmNodes(hosts=[nodename], features=features)
                partitions2features2nodes[partition_name][fl] = nodes
                partition.nodes.append(nodes)
            else:
                if nodename not in nodes.hosts:
                    nodes.hosts.append(nodename)

    for line in input.readlines():
        if match := re_nodename.search(line):
            if nodename:
                process()
            nodename = match.group(1)
        elif match := re_features.search(line):
            features = match.group(1).split(",")
        elif match := re_partitions.search(line):
            partition_names = match.group(1).split(",")

    if nodename:
        process()


def fill_partitions_configuration(input: TextIO, configuration: "SlurmConfiguration"):
    """Parses the output of scontrol show --oneliner partition"""
    re_partitionname = re.compile(r"""^PartitionName=(\w+)""")
    re_mem_per_cpu = re.compile(r"""(?:=|\s)DefMemPerCPU=(\d+)(?:\D|$)""")
    re_cpu_per_gpu = re.compile(r"""(?:=|\s)DefCpuPerGPU=(\d+)(?:\D|$)""")

    for line in input.readlines():
        if match := re_partitionname.search(line):
            name = match.group(1)
            cfg = configuration.partitions.setdefault(name, SlurmPartition(nodes=[]))

            if m := re_mem_per_cpu.search(line):
                cfg.mem_per_cpu = int(m.group(1)) * 1024

            if m := re_cpu_per_gpu.search(line):
                cfg.cpu_per_gpu = int(m.group(1))


# ---- SLURM launcher finder


def parse_size(s: Optional[str]):
    return humanfriendly.parse_size(s) if s else None


@dataclass
class GPUConfig(YAMLDataClass):
    """Represents a GPU"""

    model: Optional[str] = None
    count: int = 0
    memory: Annotated[int, Initialize(parse_size)] = 0

    min_memory: Annotated[int, Initialize(parse_size)] = 0
    """Minimum memory to be allocated on this node"""

    min_mem_ratio: Optional[float] = 0.0
    """Minimum memory ratio"""

    def update(self, other: "GPUConfig"):
        if other.model:
            self.model = other.model
        if other.count:
            self.count = other.count
        if other.memory:
            self.memory = other.memory
        if other.min_memory:
            self.min_memory = other.min_memory

    def to_spec(self):
        cuda = []
        min_memory = max(int(self.memory * self.min_mem_ratio), self.min_memory)
        cuda.extend(
            [
                CudaSpecification(self.memory, self.model, min_memory)
                for _ in range(self.count)
            ]
        )
        return cuda


@dataclass
class CPUConfig(YAMLDataClass):
    cpu_per_gpu: int = 0
    """Number of CPU per GPU"""

    mem_per_cpu: Annotated[int, Initialize(humanfriendly.parse_size)] = 0
    """Memory per CPU"""

    cores: int = 0

    memory: Annotated[int, Initialize(parse_size)] = 0

    def update(self, other: "CPUConfig"):
        if other.cpu_per_gpu:
            self.cpu_per_gpu = other.cpu_per_gpu
        if other.mem_per_cpu:
            self.mem_per_cpu = other.mem_per_cpu
        if other.memory:
            self.memory = other.memory
        if other.cores:
            self.cores = other.cores

    def to_spec(self):
        return CPUSpecification(
            memory=self.memory,
            cores=self.cores,
            mem_per_cpu=self.mem_per_cpu,
            cpu_per_gpu=self.cpu_per_gpu,
        )


@define
class SlurmHostSpecification(HostSpecification):
    features: List[str] = Factory(list)
    hosts: List[str] = Factory(list)
    partition: str = Factory(str)

    qos_id: Optional[str] = Factory(lambda: None)
    """Quality of Service"""

    account_id: Optional[str] = Factory(lambda: None)
    """Account for this host"""


@dataclass
class SlurmNodeConfiguration(YAMLDataClass):
    max_duration: Annotated[int, Initialize(humanfriendly.parse_timespan)] = 0
    """Maximum duration of a job"""

    gpu: GPUConfig = field(default_factory=GPUConfig)
    """GPU Configuration"""

    cpu: CPUConfig = field(default_factory=CPUConfig)
    """CPU Configuration"""

    def update(self, other: "SlurmNodeConfiguration"):
        if other.max_duration:
            self.max_duration = other.max_duration

        if other.gpu:
            self.gpu.update(other.gpu)

        if other.cpu:
            self.cpu.update(other.cpu)

    def to_host_spec(self):
        spec = SlurmHostSpecification(
            cpu=(self.cpu or CPUConfig()).to_spec(),
            cuda=(self.gpu or GPUConfig()).to_spec(),
        )
        spec.max_duration = self.max_duration or 0
        return spec


@dataclass
class SlurmNodes(YAMLDataClass):
    features: List[str] = field(default_factory=list)
    """Nodes features"""

    hosts: List[str] = field(default_factory=list)
    """List of hostnames"""

    configuration: Optional[SlurmNodeConfiguration] = None
    """(optional) nodes configuration"""

    count: int = 0
    """Number of hosts (if list of hosts is empty)"""


@dataclass
class SlurmPartition(YAMLDataClass):
    """A Slurm partition"""

    accounts: List[str] = field(default_factory=list)
    """List of accounts for this partition with the associated priority modifier"""

    qos: List[str] = field(default_factory=list)
    """List of QoS for this partition with the associated priority modifier"""

    nodes: List[SlurmNodes] = field(default_factory=list)
    """List of nodes"""

    configuration: Optional[SlurmNodeConfiguration] = None
    """Partition configuration"""

    priority: int = 0
    """Priority for choosing this partition (higher preferred)"""

    disabled: bool = False
    """Can be used to disable a partition"""


class FeatureConjunction(List[str]):
    def __init__(self, features: List[str]):
        super().__init__(sorted(features))

    def __hash__(self) -> int:
        return sum([hash(tag) for tag in self])


@dataclass
class SlurmFeature(YAMLDataClass):
    """Associate a configuration with a Slurm feature"""

    configuration: Optional[SlurmNodeConfiguration] = None


@dataclass
class SlurmQOS(YAMLDataClass):
    max_duration: Annotated[int, Initialize(humanfriendly.parse_timespan)] = 0
    """Maximum duration of a job"""

    min_gpu: int = 0
    """Minimum number of GPUs"""

    priority: int = 0
    """Priority modifier for this QoS"""


class NodesSpecComputer:
    def __init__(self, config: "SlurmConfiguration", partition: SlurmPartition):
        self.config = config
        self.partition = partition
        self.main_config = config
        self.config = deepcopy(config.configuration)

        self.config.max_duration = (
            self.partition.configuration.max_duration
            if self.partition.configuration
            else None
        )
        self.priority = partition.priority
        self.qos_id = None
        self.min_gpu = 0

    def update(self, config: SlurmNodeConfiguration):
        self.config = deepcopy(self.config)
        self.config.update(config)

    def update_with_qos(self, qos_id: str):
        self.qos_id = qos_id
        if qos := self.main_config.qos.get(qos_id, None):
            self.priority += qos.priority
            self.min_gpu = qos.min_gpu
            if qos.max_duration > 0:
                self.config.max_duration = qos.max_duration

    def get_host(self) -> SlurmHostSpecification:
        host = self.config.to_host_spec()
        host.priority = self.priority
        host.qos_id = self.qos_id
        host.min_gpu = self.min_gpu

        return host


class FeatureBooleanFormula:
    clauses: Set[List[str]]

    def __init__(self):
        self.clauses = set()

    def add(self, features: List[str]):
        """Adds conjunction of tags"""
        self.clauses.add(FeatureConjunction(features))

    def to_constraint(self):
        """Returns a constraint for sbatch/srun"""
        it = ("&".join(clause) for clause in self.clauses)
        s = f"""({")|(".join(it)})"""
        return None if s == "()" else s


class MatchingSpec:
    def __init__(self):
        self.fbf = FeatureBooleanFormula()
        self.hosts: set[str] = set()
        self.partitions: Set[str] = set()
        self.qos: Optional[str] = None
        self.account: Optional[str] = None
        self.mem_per_cpu: int = 0

    def update(self, host_spec: SlurmHostSpecification):
        if host_spec.qos_id != self.qos and self.qos is not None:
            # Cannot update with other QoS
            return
        self.qos = host_spec.qos_id

        if host_spec.account_id != self.account and self.account is not None:
            # Cannot update with other account
            return

        if (
            host_spec.cpu.mem_per_cpu > 0
            and self.mem_per_cpu > 0
            and host_spec.cpu.mem_per_cpu != self.mem_per_cpu
        ):
            # Cannot update with different mem per cpu
            return

        if host_spec.cpu.mem_per_cpu:
            self.mem_per_cpu = host_spec.cpu.mem_per_cpu

        self.account = host_spec.account_id

        self.partitions.add(host_spec.partition)
        self.fbf.add(host_spec.features)
        if host_spec.hosts:
            self.hosts.update(host_spec.hosts)


@dataclass
class SlurmConfiguration(YAMLDataClass, LauncherConfiguration):
    id: str
    """Slurm ID"""

    partitions: Dict[str, SlurmPartition]
    """List of partitions"""

    connector: str = "local"
    """Name of the connector"""

    path: str = "/usr/bin"
    """Path for SLURM commands"""

    use_features: bool = True
    """Whether features should be used"""

    use_hosts: bool = True
    """Whether hosts should be used in the query"""

    use_memory_contraint: bool = True
    """Whether memory constraint can be specified"""

    query_slurm: bool = False
    """True to query SLURM directly (using scontrol)"""

    tags: List[str] = field(default_factory=list)

    weight: int = 0

    qos: Dict[str, SlurmQOS] = field(default_factory=lambda: {})

    features_regex: Annotated[
        List[re.Pattern],
        Initialize(lambda regexps: [re.compile(regex) for regex in regexps]),
    ] = field(default_factory=list)
    """
    Regex to get the information from features
        - CUDA: cuda:count, cuda:memory
    """

    features: Dict[str, SlurmFeature] = field(default_factory=lambda: {})
    """List of features with associated configurations"""

    configuration: Optional[SlurmNodeConfiguration] = None
    """Partition configuration"""

    def compute(self, registry: "LauncherRegistry"):
        if self.query_slurm:
            self.query_slurm = False

            # Read node information
            connector = registry.getConnector(self.connector)
            pb = connector.processbuilder()
            pb.command = ["scontrol", "--hide", "show", "nodes"]

            def handle_output(input: io.BytesIO):
                StreamReader = codecs.getreader("utf-8")
                fill_nodes_configuration(StreamReader(input), self)

            pb.stdout = Redirect.pipe(handle_output)
            pb.start()

            # Read partition information
            pb = connector.processbuilder()
            pb.command = ["scontrol", "--hide", "show", "--oneliner", "partition"]

            def handle_output(input: io.BytesIO):
                StreamReader = codecs.getreader("utf-8")
                fill_partitions_configuration(StreamReader(input), self)

            pb.stdout = Redirect.pipe(handle_output)
            pb.start()

    @cached_property
    def computed_nodes(self) -> List[SlurmHostSpecification]:
        """Computes the list of potential compute nodes (grouped by similar nodes)"""
        hosts = []

        for partition_name, partition in self.partitions.items():
            if partition.disabled:
                continue

            for node in partition.nodes:
                nodes_spec = NodesSpecComputer(self, partition)
                nodes_spec.update(self.configuration)

                # Set partition GPU
                if partition.configuration:
                    nodes_spec.update(partition.configuration)

                if node.configuration:
                    nodes_spec.update(node.configuration)

                for feature in node.features:
                    # Use feature data directly
                    if data := self.features.get(feature, None):
                        nodes_spec.update(data.configuration)

                    # logger.debug("Looking at %s", self.features_regex)
                    for regex in self.features_regex:
                        # logger.debug("%s/%s => %s", regex, tag, regex.match(tag))
                        if m := regex.match(feature):
                            d = m.groupdict()
                            if _count := d.get("cuda_count", None):
                                nodes_spec.config.gpu.count = int(_count)
                            if memory := d.get("cuda_memory", None):
                                nodes_spec.config.gpu.memory = humanfriendly.parse_size(
                                    memory
                                )

                qos_list = partition.qos or [None]
                accounts = partition.accounts or [None]
                for qos in qos_list:
                    qos_nodes_spec = deepcopy(nodes_spec)
                    qos_nodes_spec.update_with_qos(qos)

                    host = qos_nodes_spec.get_host()
                    host.features = node.features
                    host.partition = partition_name
                    host.hosts = node.hosts

                    for account in accounts:
                        account_host = deepcopy(host)
                        account_host.account_id = account
                        hosts.append(account_host)
                    logging.debug("Computed slurm host: %s", host)

        hosts.sort(key=lambda host: -host.priority)
        return hosts

    def get(
        self, registry: "LauncherRegistry", requirement: HostRequirement
    ) -> Optional["Launcher"]:
        # Compute the configuration if needed
        self.compute(registry)

        # Compute tags or hosts

        # Current set of constraints
        current_match = None
        matching_spec = MatchingSpec()

        for node in self.computed_nodes:
            if match := requirement.match(node):
                logger.debug("Match %s for %s", match, node)

                # If score is below the current one, goes to the next one
                if current_match and (
                    match.score <= current_match.score
                    and match.requirement is not current_match.requirement
                ):
                    continue

                # If the requirement has changed, clear everything
                if not current_match or (
                    match.requirement is not current_match.requirement
                ):
                    # Clear if the requirement changed
                    logger.debug("Clearing %s / %s", current_match, match)
                    matching_spec = MatchingSpec()
                    current_match = match

                logger.debug(
                    "Adding %s, %s, %s", node.partition, node.features, node.hosts
                )
                matching_spec.update(node)

        # Returns the appropriate launcher (if any)
        use_features = matching_spec.fbf.clauses and self.use_features
        if use_features or matching_spec.hosts:
            assert current_match is not None

            # Launching using tags
            from .base import SlurmLauncher

            launcher = SlurmLauncher(
                connector=registry.getConnector(self.connector), binpath=self.path
            )

            launcher.options.partition = ",".join(matching_spec.partitions)
            launcher.options.gpus_per_node = (
                len(current_match.requirement.cuda_gpus)
                if current_match.requirement.cuda_gpus
                else None
            )

            launcher.options.qos = matching_spec.qos
            launcher.options.account = matching_spec.account

            if current_match.requirement.cpu.cores > 0:
                launcher.options.cpus_per_task = current_match.requirement.cpu.cores

            if current_match.requirement.cpu.memory > 0:
                if self.use_memory_contraint:
                    launcher.options.mem = (
                        f"{current_match.requirement.cpu.memory // (1024*1024)}M"
                    )
                else:
                    assert (
                        matching_spec.mem_per_cpu > 0
                    ), "Memory per CPU should be specified"
                    cpus_per_task = math.ceil(
                        current_match.requirement.cpu.memory / matching_spec.mem_per_cpu
                    )
                    launcher.options.cpus_per_task = max(
                        launcher.options.cpus_per_task, cpus_per_task
                    )

            if use_features:
                launcher.options.constraint = matching_spec.fbf.to_constraint()
            else:
                logger.warning("Selecting first host")
                launcher.options.nodelist = next(iter(matching_spec.hosts))

            if current_match.requirement.duration > 0:
                total_seconds = current_match.requirement.duration
                seconds = total_seconds % 60
                minutes = (total_seconds // 60) % 60
                hours = total_seconds // 3600
                launcher.options.time = f"{hours}:{minutes}:{seconds}"

            logger.debug("Slurm options: %s", " ".join(launcher.options.args()))
            return launcher

        return None
